import argparse
import gzip
import os
import signal
import sys
import traceback
import typing

import evfl
from evfl import EventFlow
import eventeditor.ai as ai
import eventeditor.actor_json as aj
from eventeditor.actor_view import ActorView
from eventeditor.event_view import EventView
from eventeditor.flow_data import FlowData, FlowDataChangeReason
from eventeditor.flowchart_view import FlowchartView
import eventeditor.util as util
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtGui as qg # type: ignore
import PyQt5.QtWidgets as q # type: ignore
from . import _version

class MainWindow(q.QMainWindow):
    def __init__(self, args) -> None:
        super().__init__()
        self.args = args
        self.flow: typing.Optional[EventFlow] = None
        self.flow_data = FlowData()
        self.flow_path = ''
        self.unsaved = False

        self.initMenu()
        self.initWidgets()
        self.initLayout()

        self.connectWidgets()
        self.centralWidget().setHidden(True)
        self.updateTitleAndActions()

        self.readSettings()

        self.initVersionInfo()

    def initVersionInfo(self) -> None:
        versions = _version.get_versions()
        self._version: str = versions['version']
        self._version_rev: str = versions['full-revisionid']

    def show(self) -> None:
        super().show()
        if self.args.event_flow_file:
            self.readFlow(self.args.event_flow_file)

    def initMenu(self) -> None:
        menu = self.menuBar()

        file_menu = menu.addMenu('&File')
        self.new_action = q.QAction('&New...', self)
        self.new_action.setShortcut(qg.QKeySequence.New)
        self.new_action.triggered.connect(self.onNewFile)
        file_menu.addAction(self.new_action)
        self.open_action = q.QAction('&Open...', self)
        self.open_action.setShortcut(qg.QKeySequence.Open)
        self.open_action.triggered.connect(self.onOpenFile)
        file_menu.addAction(self.open_action)
        file_menu.addSeparator()
        self.open_autosave_action = q.QAction('Open autosave...', self)
        self.open_autosave_action.triggered.connect(lambda: self.onOpenFile(str(self.flow_data.auto_save.get_directory()), name_filter=f'Flowchart autosave (autosave_{self.flow_data.flow.name}_*.bfevfl.gz)'))
        if not self.flow_data.auto_save.get_directory():
            self.open_autosave_action.setVisible(False)
        file_menu.addAction(self.open_autosave_action)
        self.save_action = q.QAction('&Save', self)
        self.save_action.setShortcut(qg.QKeySequence.Save)
        self.save_action.setEnabled(False)
        self.save_action.triggered.connect(self.onSaveFile)
        file_menu.addAction(self.save_action)
        self.save_as_action = q.QAction('Save as...', self)
        # No shortcut is assigned for Windows in QKeySequence.SaveAs
        # self.save_as_action.setShortcut(qg.QKeySequence.SaveAs)
        self.save_as_action.setShortcut('Ctrl+Shift+S')
        self.save_as_action.setEnabled(False)
        self.save_as_action.triggered.connect(self.onSaveAsFile)
        file_menu.addAction(self.save_as_action)
        self.rename_flow_action = q.QAction('Rename flow', self)
        self.rename_flow_action.triggered.connect(self.renameFlow)
        file_menu.addAction(self.rename_flow_action)
        file_menu.addSeparator()
        self.exit_action = q.QAction('E&xit', self)
        self.exit_action.setShortcut(qg.QKeySequence.Quit)
        self.exit_action.triggered.connect(self.close)
        file_menu.addAction(self.exit_action)

        view_menu = menu.addMenu('Flowc&hart')
        self.event_name_visible_action = q.QAction('&Show event names', self)
        self.event_name_visible_action.setCheckable(True)
        self.event_name_visible_action.setChecked(False)
        self.event_name_visible_action.triggered.connect(self.onEventNameVisibilityChanged)
        view_menu.addAction(self.event_name_visible_action)
        self.event_param_visible_action = q.QAction('&Show event parameters', self)
        self.event_param_visible_action.setCheckable(True)
        self.event_param_visible_action.setChecked(False)
        self.event_param_visible_action.triggered.connect(self.onEventParamVisibilityChanged)
        view_menu.addAction(self.event_param_visible_action)
        view_menu.addSeparator()
        self.reload_graph_action = q.QAction('&Reload graph', self)
        self.reload_graph_action.setShortcut('Ctrl+Shift+R')
        view_menu.addAction(self.reload_graph_action)
        self.export_graph_action = q.QAction('E&xport graph data to JSON...', self)
        view_menu.addAction(self.export_graph_action)
        self.export_definitions_action = q.QAction('Ex&port actor definition data to JSON...', self)
        view_menu.addAction(self.export_definitions_action)
        self.reorder_event_parameters_action = q.QAction('Reorder event parameters', self)
        view_menu.addAction(self.reorder_event_parameters_action)
        view_menu.addSeparator()
        self.add_event_action = q.QAction('&Add event...', self)
        view_menu.addAction(self.add_event_action)
        self.add_fork_action = q.QAction('Add fork...', self)
        view_menu.addAction(self.add_fork_action)

        help_menu = menu.addMenu('&Help')
        wiki_action = q.QAction('Wiki', self)
        wiki_action.triggered.connect(lambda: qg.QDesktopServices.openUrl(qc.QUrl('https://zeldamods.org')))
        help_menu.addAction(wiki_action)
        github_repo_action = q.QAction('GitHub repository', self)
        github_repo_action.triggered.connect(lambda: qg.QDesktopServices.openUrl(qc.QUrl('https://github.com/leoetlino/event-editor')))
        help_menu.addAction(github_repo_action)
        help_menu.addSeparator()
        about_action = q.QAction('About', self)
        about_action.triggered.connect(self.about)
        help_menu.addAction(about_action)

    def about(self) -> None:
        q.QMessageBox.about(self, 'About EventEditor', f'<h2>EventEditor</h2><p>EventEditor is an open-source event flow editor for <i>The Legend of Zelda: Breath of the Wild.</i></p><p><small>Version: {self._version}<br>Revision: {self._version_rev}</small></p>')
        pass

    def initWidgets(self) -> None:
        self.tab_widget = q.QTabWidget(self)
        self.tab_widget.setTabPosition(q.QTabWidget.South)

        self.flowchart_view = FlowchartView(self, self.flow_data)
        self.actor_view = ActorView(self, self.flow_data)
        self.event_view = EventView(self, self.flow_data)

    def initLayout(self) -> None:
        self.tab_widget.addTab(self.flowchart_view, 'F&lowchart')
        self.tab_widget.addTab(self.actor_view, '&Actors')
        self.tab_widget.addTab(self.event_view, '&Events')

        self.setCentralWidget(self.tab_widget)

    def connectWidgets(self) -> None:
        def set_unsaved_flag():
            self.unsaved = True
        self.flow_data.flowDataChanged.connect(lambda reason: set_unsaved_flag())
        self.flow_data.flowDataChanged.connect(lambda reason: self.updateTitleAndActions())

        self.flowchart_view.readySignal.connect(self.onViewReady)
        self.flowchart_view.eventSelected.connect(self.onEventSelected)
        self.reload_graph_action.triggered.connect(self.flowchart_view.reload)
        self.export_graph_action.triggered.connect(self.flowchart_view.export)
        self.export_definitions_action.triggered.connect(self.flowchart_view.export_definitions)
        self.reorder_event_parameters_action.triggered.connect(self.flowchart_view.reorder_event_parameters)
        self.add_event_action.triggered.connect(self.flowchart_view.addNewEvent)
        self.add_fork_action.triggered.connect(self.flowchart_view.addFork)

        self.actor_view.detail_pane.jumpToEventsRequested.connect(self.onJumpToEventsRequested)
        self.actor_view.jumpToActorEventsRequested.connect(self.onJumpToEventsRequested)
        self.event_view.jumpToFlowchartRequested.connect(self.onJumpToFlowchartRequested)

        self.tab_widget.currentChanged.connect(self.onTabChanged)

    def closeEvent(self, event) -> None:
        if not self.unsaved or not self.flow:
            event.accept()
            self.writeSettings()
            return

        ret = q.QMessageBox.question(self, 'Unsaved changes', f'{self.flow.name} has unsaved changes. Save changes before closing?', q.QMessageBox.Yes | q.QMessageBox.No | q.QMessageBox.Cancel)

        if ret == q.QMessageBox.Yes:
            self.writeFlow(self.flow_path)
            self.writeSettings()
            event.accept()
        elif ret == q.QMessageBox.No:
            self.writeSettings()
            event.accept()
        else:
            event.ignore()

    def readSettings(self) -> None:
        settings = qc.QSettings()
        ai.set_rom_path(settings.value('paths/rom_root'))
        aj.set_actor_definitions_path(settings.value('paths/actor_definitions_root'))
        settings.beginGroup('MainWindow')
        self.resize(settings.value('size', qc.QSize(800, 600)))
        self.move(settings.value('pos', qc.QPoint(200, 200)))
        settings.endGroup()

        settings.beginGroup('flowchart')
        self.event_name_visible_action.setChecked(settings.value('visible_names', False, type=bool))
        self.event_param_visible_action.setChecked(settings.value('visible_params', False, type=bool))
        settings.endGroup()

    def writeSettings(self) -> None:
        settings = qc.QSettings()
        settings.beginGroup('MainWindow')
        settings.setValue('size', self.size())
        settings.setValue('pos', self.pos())
        settings.endGroup()

        settings.beginGroup('flowchart')
        settings.setValue('visible_names', self.event_name_visible_action.isChecked())
        settings.setValue('visible_params', self.event_param_visible_action.isChecked())
        settings.endGroup()

        if aj._actor_definitions_path:
            settings.beginGroup('paths')
            settings.setValue('actor_definitions_root', str(aj._actor_definitions_path))
            settings.endGroup()

    def updateTitleAndActions(self) -> None:
        if not self.flow:
            self.setWindowTitle('EventEditor')
        else:
            indicator = '*' if self.unsaved else ''
            self.setWindowTitle(f'EventEditor - {indicator}{self.flow.name}')

        self.open_autosave_action.setEnabled(bool(self.flow) and bool(self.flow_path))
        self.save_action.setEnabled(bool(self.flow) and bool(self.flow_path))
        self.save_as_action.setEnabled(bool(self.flow))
        self.rename_flow_action.setEnabled(bool(self.flow) and bool(self.flow_path))

        self.reload_graph_action.setEnabled(bool(self.flow) and bool(self.flow_path))
        self.export_graph_action.setEnabled(bool(self.flow))
        self.export_definitions_action.setEnabled(bool(self.flow))
        self.reorder_event_parameters_action.setEnabled(bool(self.flow))
        self.add_event_action.setEnabled(bool(self.flow) and bool(self.flow_path))
        self.add_fork_action.setEnabled(bool(self.flow) and bool(self.flow_path))

    def renameFlow(self) -> None:
        if not self.flow or not self.flow.flowchart:
            return
        text, ok = q.QInputDialog.getText(self, 'Rename', 'Enter a new name for the flowchart.', q.QLineEdit.Normal, self.flow.name)
        if not ok or not text:
            return
        self.flow.name = text
        self.flow.flowchart.name = text
        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.EventFlowRename)

    def readFlow(self, path: str) -> bool:
        if self.flow and self.unsaved:
            ret = q.QMessageBox.question(self, 'Unsaved changes', f'{self.flow.name} has unsaved changes. Save changes before opening another file?', q.QMessageBox.Yes | q.QMessageBox.No | q.QMessageBox.Cancel)
            if ret == q.QMessageBox.Yes:
                self.writeFlow(self.flow_path)
            elif ret == q.QMessageBox.Cancel:
                return False

        try:
            flow = EventFlow()
            util.read_flow(path, flow)
            self.flow = flow
            self.flow_path = path
            self.flow_data.setFlow(flow)
            self.unsaved = False
            self.updateTitleAndActions()
            return True
        except:
            traceback.print_exc()
            q.QMessageBox.critical(self, 'Open', 'Failed to load event flow')
            return False

    def writeFlow(self, path: str) -> bool:
        if not self.flow or not path:
            return False

        try:
            util.write_flow(path, self.flow)
            self.flow_path = path
            self.unsaved = False
            self.updateTitleAndActions()
            return True
        except:
            traceback.print_exc()
            q.QMessageBox.critical(self, 'Save', 'Failed to write event flow. Please ensure there are no placeholder events left.')
            return False

    def onNewFile(self) -> bool:
        path = q.QFileDialog.getSaveFileName(self, 'Select a location for the new file', '', 'Flowchart (*.bfevfl)')[0]
        if not path:
            return False
        flow = evfl.EventFlow()
        flow.name = 'NewFile'
        flow.flowchart = evfl.Flowchart()
        flow.flowchart.name = 'NewFile'
        try:
            util.write_flow(path, flow)
        except:
            traceback.print_exc()
            q.QMessageBox.critical(self, 'New file', 'Failed to write new event flow -- cannot continue')
            return False
        return self.readFlow(path)

    def onOpenFile(self, default_directory='', name_filter='Flowchart (*.bfevfl)') -> bool:
        default_directory_ = default_directory if default_directory else self.flow_path
        path = q.QFileDialog.getOpenFileName(self, 'Open event flowchart', default_directory_, name_filter)[0]
        if path:
            return self.readFlow(path)
        return False

    def onSaveFile(self) -> None:
        self.writeFlow(self.flow_path)

    def onSaveAsFile(self) -> None:
        path = q.QFileDialog.getSaveFileName(self, 'Save as...', '', 'Flowchart (*.bfevfl)')[0]
        self.writeFlow(path)

    def onTabChanged(self, idx: int) -> None:
        self.flowchart_view.setIsCurrentView(self.tab_widget.widget(idx) == self.flowchart_view)

    def onViewReady(self) -> None:
        self.centralWidget().setHidden(False)
        self.onEventNameVisibilityChanged()
        self.onEventParamVisibilityChanged()

    def onEventSelected(self, event_idx: int) -> None:
        self.event_view.selectEvent(event_idx)

    def onJumpToEventsRequested(self, filter_str: str = '') -> None:
        self.tab_widget.setCurrentWidget(self.event_view)
        if filter_str:
            self.event_view.search_bar.setValue(filter_str)
            self.event_view.search_bar.show()

    def onJumpToFlowchartRequested(self, idx: int) -> None:
        """Request a node select in the flowchart webview. Negative indices are used for entry points."""
        self.tab_widget.setCurrentWidget(self.flowchart_view)
        self.flowchart_view.selectRequested.emit(idx)

    def onEventNameVisibilityChanged(self) -> None:
        visible = self.event_name_visible_action.isChecked()
        self.flowchart_view.eventNameVisibilityChanged.emit(visible)

    def onEventParamVisibilityChanged(self) -> None:
        visible = self.event_param_visible_action.isChecked()
        self.flowchart_view.eventParamVisibilityChanged.emit(visible)

def main() -> None:
    qc.QCoreApplication.setOrganizationName('eventeditor')
    qc.QCoreApplication.setApplicationName('eventeditor')
    qc.QSettings.setDefaultFormat(qc.QSettings.IniFormat)

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    parser = argparse.ArgumentParser(prog='eventeditor', description='An event editor for Breath of the Wild')
    parser.add_argument('event_flow_file', nargs='?', help='Event flow file to open')
    args, _ = parser.parse_known_args()
    app = q.QApplication(sys.argv)
    if os.name == 'nt':
        app_font = app.font()
        app_font.setFamily('Segoe UI')
        app_font.setPointSize(int(qg.QFontInfo(app_font).pointSize() * 1.20))
        app.setFont(app_font)
    win = MainWindow(args)
    win.show()
    ret = app.exec_()
    sys.exit(ret)

if __name__ == '__main__':
    main()



from enum import IntEnum, auto
import typing

from evfl import EventFlow, Actor, ActorIdentifier
from evfl.entry_point import EntryPoint
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtWidgets as q # type: ignore
import chatgpt_events as cge

class ActorModelColumn(IntEnum):
    Name = 0
    SubName = auto()
    ArgumentName = auto()
    ArgumentEntryPoint = auto()
    NumActions = auto()
    NumQueries = auto()

class ActorModel(qc.QAbstractTableModel):
    def __init__(self, *kwargs) -> None:
        super().__init__(*kwargs)
        self.flow: typing.Optional[EventFlow] = None
        self.l: typing.List[Actor] = []

    def refresh(self) -> None:
        self.dataChanged.emit(self.createIndex(0, 0),
                              self.createIndex(self.rowCount(None) - 1, self.columnCount(None) - 1))

    def has(self, identifier: ActorIdentifier) -> bool:
        return any(actor.identifier == identifier for actor in self.l)

    def setData(self, index: qc.QModelIndex, value, role) -> bool:
        if role != qc.Qt.EditRole:
            return False
        row = index.row()
        col = index.column()
        actor = self.l[row]

        if col == ActorModelColumn.Name:
            actor.identifier.name = value
            # Optionally regenerate the description when the name changes
            actor.description = cge.generate_text(f"Update the description for actor named {value}")
        elif col == ActorModelColumn.SubName:
            actor.identifier.sub_name = value
        elif col == ActorModelColumn.ArgumentName:
            actor.argument_name = value
        elif col == ActorModelColumn.ArgumentEntryPoint:
            actor.argument_entry_point.v = value

        self.dataChanged.emit(index, index)
        return True

    def appendEmptyActor(self) -> bool:
        if not self.flow or not self.flow.flowchart:
            return False
        self.beginInsertRows(qc.QModelIndex(), len(self.l), len(self.l))
        new_actor = Actor()
        # Generate an initial description for the new actor
        description = cge.generate_text("Generate a description for a new actor")
        new_actor.description = description  # Assuming Actor class has a description attribute
        self.l.append(new_actor)
        self.endInsertRows()
        return True

    def remove(self, actor: Actor) -> bool:
        if not self.flow or not self.flow.flowchart:
            return False
        row = self.l.index(actor)
        if row is None:
            return False
        self.beginRemoveRows(qc.QModelIndex(), row, row)
        self.l.pop(row)
        self.endRemoveRows()
        return True

    def columnCount(self, parent) -> int:
        return len(ActorModelColumn)

    def rowCount(self, parent) -> int:
        return len(self.l)

    def flags(self, index: qc.QModelIndex) -> qc.Qt.ItemFlags:
        if index.column() > ActorModelColumn.ArgumentEntryPoint:
            return super().flags(index)
        return qc.Qt.ItemIsEditable | super().flags(index)

    def setData(self, index: qc.QModelIndex, value, role) -> bool:
        if role != qc.Qt.EditRole:
            return False
        row = index.row()
        col = index.column()
        if col == ActorModelColumn.Name:
            assert isinstance(value, str)
            self.l[row].identifier.name = value
        if col == ActorModelColumn.SubName:
            assert isinstance(value, str)
            self.l[row].identifier.sub_name = value
        if col == ActorModelColumn.ArgumentName:
            assert isinstance(value, str)
            self.l[row].argument_name = value
        if col == ActorModelColumn.ArgumentEntryPoint:
            assert isinstance(value, EntryPoint) or value is None
            self.l[row].argument_entry_point.v = value
        self.dataChanged.emit(index, index)
        return True

    def headerData(self, section, orientation, role) -> qc.QVariant:
        if role != qc.Qt.DisplayRole:
            return qc.QVariant()

        if section == ActorModelColumn.Name:
            return 'Name'
        if section == ActorModelColumn.SubName:
            return 'Sub name'
        if section == ActorModelColumn.ArgumentName:
            return 'Argument name'
        if section == ActorModelColumn.ArgumentEntryPoint:
            return 'Argument entry point'
        if section == ActorModelColumn.NumActions:
            return 'Actions'
        if section == ActorModelColumn.NumQueries:
            return 'Queries'
        return 'Unknown'

    def data(self, index, role) -> qc.QVariant:
        col = index.column()
        row = index.row()

        if role == qc.Qt.UserRole:
            return self.l[row]

        if role == qc.Qt.EditRole:
            if col == ActorModelColumn.Name:
                return self.l[row].identifier.name
            if col == ActorModelColumn.SubName:
                return self.l[row].identifier.sub_name
            if col == ActorModelColumn.ArgumentName:
                return self.l[row].argument_name
            if col == ActorModelColumn.ArgumentEntryPoint:
                return self.l[row].argument_entry_point.v
            return qc.QVariant()

        if role != qc.Qt.DisplayRole and role != qc.Qt.ToolTipRole:
            return qc.QVariant()

        if col == ActorModelColumn.Name:
            return self.l[row].identifier.name
        if col == ActorModelColumn.SubName:
            return self.l[row].identifier.sub_name or '–'
        if col == ActorModelColumn.ArgumentName:
            return self.l[row].argument_name or '–'
        if col == ActorModelColumn.ArgumentEntryPoint:
            if not self.l[row].argument_entry_point.v:
                return '–'
            return str(self.l[row].argument_entry_point.v.name)
        if col == ActorModelColumn.NumActions:
            return str(len(self.l[row].actions))
        if col == ActorModelColumn.NumQueries:
            return str(len(self.l[row].queries))
        return qc.QVariant()



import json
import typing

import eventeditor.actor_json as aj
from eventeditor.container_model import ContainerModel
from eventeditor.container_view import ContainerView
from eventeditor.event_branch_editors import SwitchEventEditDialog, ForkEventEditDialog
from eventeditor.event_edit_dialog import show_event_editor
from eventeditor.event_chooser_dialog import show_event_type_chooser, add_new_event, EventChooserDialog, CheckableEventParentListWidget
from eventeditor.event_fork_chooser_dialog import EventForkChooserDialog
from eventeditor.flow_data import FlowData, FlowDataChangeReason
import eventeditor.flowchart_tools as ft
from eventeditor.search_bar import SearchBar
from eventeditor.util import *
from evfl import Container, Flowchart, Actor, Event, EventFlow, ActionEvent, SwitchEvent, ForkEvent, JoinEvent, SubFlowEvent
from evfl.common import Index, RequiredIndex
from evfl.entry_point import EntryPoint
from evfl.enums import EventType
from evfl.repr_util import generate_flowchart_graph
from PyQt5.QtWebChannel import QWebChannel # type: ignore
from PyQt5.QtWebEngineWidgets import QWebEngineView # type: ignore
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtGui as qg # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class FlowchartWebObject(qc.QObject):
    flowDataChanged = qc.pyqtSignal()
    fileLoaded = qc.pyqtSignal(EventFlow)
    eventNameVisibilityChanged = qc.pyqtSignal(bool)
    eventParamVisibilityChanged = qc.pyqtSignal(bool)
    actionProhibitionChanged = qc.pyqtSignal(bool)

    selectRequested = qc.pyqtSignal(int)

    def __init__(self, view) -> None:
        super().__init__(view)
        self.view: FlowchartView = view

    @qc.pyqtSlot(result=qc.QVariant)
    def getJson(self) -> qc.QVariant:
        return qc.QVariant(json.loads(json.dumps(self.getData(), default=lambda x: str(x))))

    def getData(self) -> list:
        flow = self.view.flow_data.flow
        return generate_flowchart_graph(flow) if flow else []

    @qc.pyqtSlot()
    def emitReadySignal(self):
        self.view.readySignal.emit()

    @qc.pyqtSlot(int)
    def emitEventSelectedSignal(self, node_id: int) -> None:
        self.view.eventSelected.emit(int(node_id))

    @qc.pyqtSlot()
    def emitReloadedSignal(self):
        self.view.reloadedSignal.emit()

    @qc.pyqtSlot(int)
    def editEvent(self, node_id: int):
        self.view.webEditEvent(int(node_id))

    @qc.pyqtSlot(int)
    def addEntryPoint(self, node_id: int):
        self.view.webAddEntryPoint(int(node_id))

    @qc.pyqtSlot(int)
    def removeEntryPoint(self, node_id: int):
        self.view.webRemoveEntryPoint(-1000-int(node_id))

    @qc.pyqtSlot(list, int)
    def addEventAbove(self, parents: typing.List[int], node_id: int):
        self.view.webAddEventAbove([int(x) for x in parents], int(node_id))

    @qc.pyqtSlot(int)
    def addEventBelow(self, node_id: int):
        self.view.webAddEventBelow(int(node_id))

    @qc.pyqtSlot(int)
    def unlink(self, node_id: int):
        self.view.webUnlink(int(node_id))

    @qc.pyqtSlot(int)
    def link(self, node_id: int):
        self.view.webLink(int(node_id))

    @qc.pyqtSlot(list, int)
    def removeEvent(self, parents: typing.List[int], node_id: int):
        self.view.webRemoveEvent([int(x) for x in parents], int(node_id))

    @qc.pyqtSlot(int)
    def editSwitchBranches(self, node_id: int):
        self.view.webEditSwitchBranches(int(node_id))

    @qc.pyqtSlot(int)
    def editForkBranches(self, node_id: int):
        self.view.webEditForkBranches(int(node_id))

class FlowchartView(q.QWidget):
    selectRequested = qc.pyqtSignal(int)
    eventNameVisibilityChanged = qc.pyqtSignal(bool)
    eventParamVisibilityChanged = qc.pyqtSignal(bool)

    # View -> Core
    readySignal = qc.pyqtSignal()
    reloadedSignal = qc.pyqtSignal()
    eventSelected = qc.pyqtSignal(int)

    def __init__(self, parent, flow_data: FlowData) -> None:
        super().__init__(parent)
        self.flow_data: FlowData = flow_data
        self.is_current = True
        self.selected_event: typing.Optional[Event] = None
        self.showEventParams = False
        self.initWidgets()
        self.initLayout()
        self.connectWidgets()

    def initWidgets(self) -> None:
        self.web_object = FlowchartWebObject(self)
        self.flow_data.flowDataChanged.connect(self.onFlowDataChanged)
        self.flow_data.fileLoaded.connect(self.web_object.fileLoaded)
        self.selectRequested.connect(self.web_object.selectRequested)
        self.eventNameVisibilityChanged.connect(self.web_object.eventNameVisibilityChanged)
        self.eventParamVisibilityChanged.connect(self.onEventParamVisibilityChanged)
        self.eventParamVisibilityChanged.connect(self.web_object.eventParamVisibilityChanged)

        self.view = QWebEngineView()
        self.view.setContextMenuPolicy(qc.Qt.NoContextMenu)
        self.channel = QWebChannel()
        self.channel.registerObject('widget', self.web_object)
        self.view.page().setWebChannel(self.channel)
        self.view.page().setBackgroundColor(qg.QColor(0x38, 0x38, 0x38));
        self.view.setUrl(qc.QUrl.fromLocalFile(get_path('assets/index.html')))

        self.entry_point_view = q.QListView(self)
        self.ep_proxy_model = qc.QSortFilterProxyModel(self)
        self.ep_proxy_model.setSourceModel(self.flow_data.entry_point_model)
        self.ep_proxy_model.setFilterKeyColumn(-1)
        self.entry_point_view.setModel(self.ep_proxy_model)
        self.ep_search = SearchBar()
        self.ep_search.hide()

        self.container_model = ContainerModel(self)
        self.container_view = ContainerView(None, self.container_model, self.flow_data)
        self.container_stacked_widget = q.QStackedWidget()
        self.container_stacked_widget.addWidget(q.QWidget())
        self.container_stacked_widget.addWidget(self.container_view)

        self.update_timer = qc.QTimer(self)
        self.update_timer.timeout.connect(self.web_object.flowDataChanged)
        self.update_timer.setSingleShot(True)

    def initLayout(self) -> None:
        left_pane_splitter = q.QSplitter(qc.Qt.Vertical)
        ep_widget = q.QWidget()
        ep_layout = q.QVBoxLayout(ep_widget)
        ep_layout.setContentsMargins(0, 0, 0, 0)
        ep_layout.addWidget(self.entry_point_view, stretch=1)
        ep_layout.addWidget(self.ep_search)
        left_pane_splitter.addWidget(ep_widget)
        left_pane_splitter.addWidget(self.container_stacked_widget)
        left_pane_splitter.setSizes([int(left_pane_splitter.height() * 0.6), int(left_pane_splitter.height() * 0.4)])

        splitter = q.QSplitter()
        splitter.addWidget(left_pane_splitter)
        splitter.addWidget(self.view)
        splitter.setSizes([int(splitter.width() * 0.3), int(splitter.width() * 0.7)])
        layout = q.QHBoxLayout(self)
        layout.addWidget(splitter)
        layout.setContentsMargins(0, 0, 0, 0)

    def connectWidgets(self) -> None:
        self.ep_search.connectToFilterModel(self.ep_proxy_model)
        self.ep_search.addFindShortcut(self)

        self.flow_data.flowDataChanged.connect(lambda reason: self.entry_point_view.clearSelection())
        self.entry_point_view.selectionModel().selectionChanged.connect(self.onEntryPointSelected)

        connect_model_change_signals(self.container_model, self.flow_data, FlowDataChangeReason.EventParameters)
        self.eventSelected.connect(self.onEventSelectedInWebView)
        self.flow_data.flowDataChanged.connect(lambda reason: self.refreshParamModel())

        self.reloadedSignal.connect(self.onWebViewReloaded)

    def onEventParamVisibilityChanged(self, show: bool) -> None:
        self.showEventParams = show

    def setIsCurrentView(self, is_current: bool) -> None:
        self.is_current = is_current
        if is_current and self.update_timer.isActive():
            self.update_timer.stop()
            self.web_object.flowDataChanged.emit()

    def export(self) -> None:
        if not self.flow_data.flow:
            return
        path = q.QFileDialog.getSaveFileName(self, 'Select a location for the graph data', self.flow_data.flow.name + '.json', 'Data (*.json)')[0]
        if not path:
            return
        data = self.web_object.getData()
        try:
            with open(path, 'w') as f:
                json.dump(data, f, default=lambda x: str(x))
        except:
            q.QMessageBox.critical(self, 'Export graph data', 'Failed to write to ' + path)
    
    def export_definitions(self) -> None:
        try:
            aj.export_definitions(self.flow_data.flow, self)
        except:
            q.QMessageBox.critical(self, 'Export actor definition data', 'Failed to write to ' + str(aj._actor_definitions_path))
    
    def reorder_event_parameters(self) -> None:
        ft.reorder_event_flow_parameters(self.flow_data.flow)
        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.EventParameters)

    def reload(self) -> None:
        self.view.reload()

    def onWebViewReloaded(self) -> None:
        if not self.selected_event or not self.flow_data.flow or not self.flow_data.flow.flowchart:
            return

        try:
            new_idx = self.flow_data.flow.flowchart.events.index(self.selected_event)
            self.selectRequested.emit(new_idx)
        except ValueError:
            self.container_model.set(None)
            self.container_stacked_widget.setCurrentIndex(0)

    def refreshParamModel(self) -> bool:
        if self.selected_event and hasattr(self.selected_event.data, 'params'):
            if not self.selected_event.data.params: # type: ignore
                self.selected_event.data.params = Container() # type: ignore
            self.container_model.set(self.selected_event.data.params) # type: ignore
            self.container_stacked_widget.setCurrentIndex(1)
            return True
        return False

    def onEventSelectedInWebView(self, idx: int) -> None:
        if idx >= 0:
            event = self.flow_data.flow.flowchart.events[idx]
            self.selected_event = event
            if self.refreshParamModel():
                return
        else:
            self.selected_event = None

        self.container_model.set(None)
        self.container_stacked_widget.setCurrentIndex(0)

    def onFlowDataChanged(self, reason: FlowDataChangeReason) -> None:
        should_reload = bool(reason & (FlowDataChangeReason.Reset|FlowDataChangeReason.Actors|FlowDataChangeReason.Events))
        if self.showEventParams:
            should_reload = should_reload or bool(reason & FlowDataChangeReason.EventParameters)
        if not should_reload:
            return
        if self.is_current:
            self.web_object.flowDataChanged.emit()
        else:
            self.update_timer.start(15*1000)

    def onEntryPointSelected(self, selected, deselected) -> None:
        if len(selected.indexes()) != 1:
            return
        idx = selected.indexes()[0]
        self.selectRequested.emit(-1000-self.ep_proxy_model.mapToSource(idx).row())

    def delayedSelect(self, event: Event) -> None:
        try:
            qc.QTimer.singleShot(1000, lambda:
                self.selectRequested.emit(self.flow_data.flow.flowchart.events.index(event)))
        except ValueError:
            pass

    def webEditEvent(self, idx: int) -> None:
        if idx < 0:
            return
        show_event_editor(self, self.flow_data, idx)

    def webAddEntryPoint(self, event_idx: int) -> None:
        if event_idx < 0:
            return

        ep_name, ok = q.QInputDialog.getText(self, 'Add entry point', f'Name of the new entry point:', q.QLineEdit.Normal)
        if not ok or not ep_name:
            return

        ep = EntryPoint(ep_name)
        assert self.flow_data.flow and self.flow_data.flow.flowchart
        ep.main_event.v = self.flow_data.flow.flowchart.events[event_idx]
        self.flow_data.entry_point_model.append(ep)

    def webRemoveEntryPoint(self, ep_idx: int) -> None:
        try:
            self.flow_data.entry_point_model.removeRow(ep_idx)
        except IndexError as e:
            q.QMessageBox.critical(self, 'Bug', f'An error has occurred: {e}\n\nPlease report this issue and mention what you were doing when this message showed up.')

    def addNewEvent(self) -> typing.Optional[Event]:
        return add_new_event(self, self.flow_data)

    def webAddEventAbove(self, parent_indices: typing.List[int], event_idx: int) -> None:
        if event_idx < 0:
            return
        assert self.flow_data.flow and self.flow_data.flow.flowchart
        event = self.flow_data.flow.flowchart.events[event_idx]

        parent_events = [self.flow_data.flow.flowchart.events[i] for i in parent_indices if i >= 0]
        list_widget = CheckableEventParentListWidget(None, event, parent_events)
        if parent_events:
            dialog = q.QDialog(self, qc.Qt.WindowTitleHint | qc.Qt.WindowSystemMenuHint)
            dialog.setWindowTitle('Add new event above...')
            btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Ok | q.QDialogButtonBox.Cancel);
            btn_box.accepted.connect(dialog.accept)
            btn_box.rejected.connect(dialog.reject)
            dialog_layout = q.QVBoxLayout(dialog)
            dialog_layout.addWidget(q.QLabel('Please select links that should be modified to point to the new event you are going to add.'))
            dialog_layout.addWidget(list_widget)
            dialog_layout.addWidget(btn_box)
            ret = dialog.exec_()
            if not ret:
                return

        new_parent = self.addNewEvent()
        if not new_parent:
            return

        self._doAddEventAbove(list_widget.getSelectedEvents(), event, new_parent)
        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.Events)
        self.delayedSelect(new_parent)

    def _doAddEventAbove(self, parents: typing.List[typing.Tuple[Event, typing.List[typing.Any]]], event: Event, new_parent: Event) -> None:
        # Update the parents to point to the new parent.
        for parent, branches in parents:
            if isinstance(parent.data, ActionEvent) or isinstance(parent.data, JoinEvent) or isinstance(parent.data, SubFlowEvent):
                # Easy case: just set the next pointer to the new parent.
                parent.data.nxt.v = new_parent

            # For switch and fork events, update all branches that currently point to the event.
            elif isinstance(parent.data, SwitchEvent):
                for case in branches:
                    if parent.data.cases[case].v == event:
                        parent.data.cases[case].v = new_parent
            elif isinstance(parent.data, ForkEvent):
                for i, fork in enumerate(branches):
                    if fork.v == event:
                        parent.data.forks[i].v = new_parent

        # Make the new parent point to the event.
        if isinstance(new_parent.data, ActionEvent):
            new_parent.data.nxt.v = event
        elif isinstance(new_parent.data, SwitchEvent):
            new_parent.data.cases[-1] = RequiredIndex()
            new_parent.data.cases[-1].v = event
        elif isinstance(new_parent.data, SubFlowEvent):
            new_parent.data.nxt.v = event
        elif isinstance(new_parent.data, ForkEvent):
            new_parent.data.forks.clear()
            ri: RequiredIndex[Event] = RequiredIndex()
            ri.v = event
            new_parent.data.forks.append(ri)

    def webAddEventBelow(self, event_idx: int) -> None:
        if event_idx < 0:
            return
        assert self.flow_data.flow and self.flow_data.flow.flowchart
        event = self.flow_data.flow.flowchart.events[event_idx]
        new_event = self.addNewEvent()
        if new_event:
            self.webDoAddEventBelow(event, new_event)

    def webDoAddEventBelow(self, event: Event, target: Event) -> None:
        if not (isinstance(event.data, ActionEvent) or isinstance(event.data, SubFlowEvent) or isinstance(event.data, JoinEvent)):
            return

        if isinstance(target.data, ActionEvent):
            target.data.nxt.v = event.data.nxt.v
        elif isinstance(target.data, SwitchEvent):
            if event.data.nxt.v:
                target.data.cases[-1] = RequiredIndex()
                target.data.cases[-1].v = event.data.nxt.v
        elif isinstance(target.data, SubFlowEvent):
            target.data.nxt.v = event.data.nxt.v

        event.data.nxt.v = target

        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.Events)
        self.delayedSelect(target)

    def webLink(self, event_idx: int) -> None:
        if event_idx < 0:
            return
        assert self.flow_data.flow and self.flow_data.flow.flowchart
        event = self.flow_data.flow.flowchart.events[event_idx]
        self.web_object.actionProhibitionChanged.emit(True)
        dialog = EventChooserDialog(self, self.flow_data)
        dialog.event_view.jumpToFlowchartRequested.connect(self.selectRequested)
        self.eventSelected.connect(dialog.event_view.selectEvent)
        dialog.accepted.connect(lambda: self.webDoLink(event, dialog.getSelectedEvent()))
        dialog.finished.connect(lambda: self.web_object.actionProhibitionChanged.emit(False))
        dialog.show()

    def webDoLink(self, event: Event, target: Event) -> None:
        if event == target:
            q.QMessageBox.critical(self, 'Invalid choice', 'Cannot link an event to itself. Please choose another event and try again.')
            return
        event.data.nxt.v = target # type: ignore
        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.Events)
        self.delayedSelect(target)

    def webUnlink(self, event_idx: int) -> None:
        ret = q.QMessageBox.question(self, 'Unlink', 'Warning: Unlinking events that are in nested fork branches can currently result in graph corruption. Continue?')
        if ret != q.QMessageBox.Yes:
            return

        assert self.flow_data.flow and self.flow_data.flow.flowchart
        event = self.flow_data.flow.flowchart.events[event_idx]
        event.data.nxt.v = None # type: ignore
        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.Events)
        self.delayedSelect(event)

    def _findForkEventLeafNodes(self, starting_event: Event) -> typing.List[Event]:
        assert isinstance(starting_event.data, ForkEvent)
        parents: typing.List[Event] = []
        visited: typing.Set[Event] = set()

        def handleNextEvent(event: Event, next_event: typing.Optional[Event], join_stack: typing.List[Event]) -> None:
            if not next_event:
                if not join_stack:
                    parents.append(event)
                return
            traverse(next_event, join_stack)

        def traverse(event: Event, join_stack: typing.List[Event]) -> None:
            if event in visited:
                return
            visited.add(event)
            data = event.data
            if isinstance(data, ActionEvent):
                handleNextEvent(event, data.nxt.v, join_stack)
            elif isinstance(data, SwitchEvent):
                for value, case in data.cases.items():
                    traverse(case.v, join_stack)
            elif isinstance(data, ForkEvent):
                join_stack.append(data.join.v)
                for fork in data.forks:
                    traverse(fork.v, join_stack)
                traverse(data.join.v, join_stack)
            elif isinstance(data, JoinEvent):
                join_stack.pop()
                handleNextEvent(event, data.nxt.v, join_stack)
            elif isinstance(data, SubFlowEvent):
                handleNextEvent(event, data.nxt.v, join_stack)

        for fork in starting_event.data.forks:
            traverse(fork.v, [])
        return parents

    def _doRemoveEvent(self, parents: typing.List[Event], event_idx: int) -> None:
        """Erase an event from the tree, ensuring that the next pointers of parents are updated.

        This does not resize the event list. None items must be removed from the list afterwards."""
        assert self.flow_data.flow and self.flow_data.flow.flowchart
        event = self.flow_data.flow.flowchart.events[event_idx]

        next_event: typing.Optional[Event] = None
        if isinstance(event.data, ActionEvent) or isinstance(event.data, JoinEvent) or isinstance(event.data, SubFlowEvent):
            next_event = event.data.nxt.v
        elif isinstance(event.data, SwitchEvent):
            next_event = next(iter(event.data.cases.values())).v if event.data.cases else None
        elif isinstance(event.data, ForkEvent):
            assert len(event.data.forks) == 1
            next_event = event.data.forks[0].v

        # Don't let the user delete the only branch in a fork event
        if len(parents) == 1 and isinstance(parents[0].data, ForkEvent) and len(parents[0].data.forks) == 1 and not next_event:
            q.QMessageBox.information(self, 'Cannot remove', 'Please delete the parent fork event first.')
            return

        # Make the parents point to the next event.
        for parent in parents:
            if isinstance(parent.data, ActionEvent) or isinstance(parent.data, JoinEvent) or isinstance(parent.data, SubFlowEvent):
                parent.data.nxt.v = next_event

            # For switch and fork events, update all branches that currently point to the event.
            # Or remove them if there is no next event.
            elif isinstance(parent.data, SwitchEvent):
                for case in list(parent.data.cases.keys()):
                    if parent.data.cases[case].v == event:
                        if next_event:
                            parent.data.cases[case].v = next_event
                        else:
                            del parent.data.cases[case]
            elif isinstance(parent.data, ForkEvent):
                new_forks = []
                for fork in parent.data.forks:
                    if fork.v != event:
                        new_forks.append(fork)
                    elif next_event:
                        ri: RequiredIndex[Event] = RequiredIndex()
                        ri.v = next_event
                        new_forks.append(ri)
                parent.data.forks = new_forks

        # If we are removing a fork event, also remove the associated join.
        if isinstance(event.data, ForkEvent):
            self._doRemoveEvent(self._findForkEventLeafNodes(event),
                self.flow_data.flow.flowchart.events.index(event.data.join.v))

        # Ensure that entry points point to the correct event.
        for entry_point in self.flow_data.flow.flowchart.entry_points:
            if entry_point.main_event.v == event:
                entry_point.main_event.v = next_event

        # Erase this event from the list. None elements will be swept by the caller.
        self.flow_data.flow.flowchart.events[event_idx] = None # type: ignore

    def webRemoveEvent(self, parent_indices: typing.List[int], event_idx: int) -> None:
        if event_idx < 0:
            return
        assert self.flow_data.flow and self.flow_data.flow.flowchart

        parents = [self.flow_data.flow.flowchart.events[idx] for idx in parent_indices if idx >= 0]
        self._doRemoveEvent(parents, event_idx)
        self.flow_data.flow.flowchart.events = \
            [event for event in self.flow_data.flow.flowchart.events if event is not None]
        # Since we're editing the array directly, a model reset MUST be triggered.
        self.flow_data.event_model.set(self.flow_data.flow)

        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.Events)

    def webEditSwitchBranches(self, event_idx: int) -> None:
        if event_idx < 0:
            return
        assert self.flow_data.flow and self.flow_data.flow.flowchart
        event = self.flow_data.flow.flowchart.events[event_idx]
        if not isinstance(event.data, SwitchEvent):
            return
        self.web_object.actionProhibitionChanged.emit(True)
        dialog = SwitchEventEditDialog(self, event.data.cases, self.flow_data)
        dialog.finished.connect(lambda: self.web_object.actionProhibitionChanged.emit(False))
        dialog.chooserEventDoubleClicked.connect(self.selectRequested)
        self.eventSelected.connect(dialog.chooserSelectSignal)
        dialog.show()

    def webEditForkBranches(self, event_idx: int) -> None:
        if event_idx < 0:
            return
        assert self.flow_data.flow and self.flow_data.flow.flowchart
        event = self.flow_data.flow.flowchart.events[event_idx]
        if not isinstance(event.data, ForkEvent):
            return
        self.web_object.actionProhibitionChanged.emit(True)
        dialog = ForkEventEditDialog(self, event.data.forks, self.flow_data)
        dialog.finished.connect(lambda: self.web_object.actionProhibitionChanged.emit(False))
        dialog.chooserEventDoubleClicked.connect(self.selectRequested)
        self.eventSelected.connect(dialog.chooserSelectSignal)
        dialog.show()

    def addFork(self) -> None:
        self.web_object.actionProhibitionChanged.emit(True)
        dialog = EventForkChooserDialog(self, self.flow_data)
        dialog.finished.connect(lambda: self.web_object.actionProhibitionChanged.emit(False))
        dialog.accepted.connect(lambda: self._doAddFork(*dialog.getEventPair()))
        dialog.chooserEventDoubleClicked.connect(self.selectRequested)
        self.eventSelected.connect(dialog.chooserSelectSignal)
        dialog.show()

    def _findEventParentNodes(self, event: Event) -> typing.List[typing.Tuple[Event, typing.List[typing.Any]]]:
        parents: typing.List[typing.Tuple[Event, typing.List[typing.Any]]] = []
        for e in self.flow_data.flow.flowchart.events:
            data = e.data
            if isinstance(data, ActionEvent) or isinstance(data, JoinEvent) or isinstance(data, SubFlowEvent):
                if data.nxt.v == event:
                    parents.append((e, []))
            elif isinstance(data, SwitchEvent):
                if any(case.v == event for case in data.cases.values()):
                    parents.append((e, list(data.cases.keys())))
            elif isinstance(data, ForkEvent):
                if any(fork.v == event for fork in data.forks):
                    parents.append((e, data.forks))
        return parents

    def _doAddFork(self, start: Event, end: Event) -> None:
        if not (isinstance(end.data, ActionEvent) or isinstance(end.data, SubFlowEvent)):
            q.QMessageBox.critical(self, 'Not implemented', 'The end event must be an action or sub flow event currently')
            return

        assert self.flow_data.flow and self.flow_data.flow.flowchart

        # Add the fork event as a parent.
        fork_event = Event()
        fork_event.name = self.flow_data.generateEventName()
        fork_event.data = ForkEvent()
        # Add the event manually and do NOT send change signals until the join event is added.
        self.flow_data.flow.flowchart.events.append(fork_event)
        self._doAddEventAbove(self._findEventParentNodes(start), start, fork_event)

        # Fix entry points.
        for entry_point in self.flow_data.flow.flowchart.entry_points:
            if entry_point.main_event.v == start:
                entry_point.main_event.v = fork_event

        # Add the join event as a child.
        join_event = Event()
        join_event.name = self.flow_data.generateEventName()
        join_event.data = JoinEvent()
        join_event.data.nxt.v = end.data.nxt.v
        end.data.nxt.v = None
        self.flow_data.flow.flowchart.events.append(join_event)
        fork_event.data.join.v = join_event

        # Trigger a full model reset since we updated the underlying array directly.
        self.flow_data.event_model.set(self.flow_data.flow)
        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.Events)



import copy
import traceback
import typing
import yaml

import eventeditor.util as util
from evfl import ActorIdentifier
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtGui as qg # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class ActorIdentifierEditDialog(q.QDialog):
    def __init__(self, parent, identifier: ActorIdentifier) -> None:
        super().__init__(parent, qc.Qt.WindowTitleHint | qc.Qt.WindowSystemMenuHint)
        self.setWindowTitle('Edit actor identifier')
        self.setMinimumWidth(500)
        self.identifier = copy.copy(identifier)

        form_layout = q.QFormLayout()
        self.name_box = q.QLineEdit()
        self.name_box.setText(identifier.name)
        form_layout.addRow('&Name:', self.name_box)
        self.sub_name_box = q.QLineEdit()
        self.sub_name_box.setText(identifier.sub_name)
        form_layout.addRow('&Sub name:', self.sub_name_box)

        layout = q.QVBoxLayout(self)
        layout.addLayout(form_layout)
        btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Save | q.QDialogButtonBox.Cancel);
        layout.addWidget(btn_box)

        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)

    def accept(self) -> None:
        self.identifier.name = self.name_box.text()
        self.identifier.sub_name = self.sub_name_box.text()
        super().accept()

class ArrayEditDialog(q.QDialog):
    def __init__(self, parent, data: list) -> None:
        super().__init__(parent, qc.Qt.WindowTitleHint | qc.Qt.WindowSystemMenuHint)
        self.resize(350, 100)
        self.data = data
        self.setWindowTitle(f'Edit {type(self.data[0]).__name__} array')

        label = q.QLabel('New array data:')
        self.tedit = q.QPlainTextEdit()
        font = qg.QFontDatabase.systemFont(qg.QFontDatabase.FixedFont)
        font.setPointSize(int(font.pointSize() * 1.05))
        self.tedit.setFont(font)
        self.tedit.setPlainText(yaml.dump(self.data, allow_unicode=True, Dumper=yaml.SafeDumper))

        layout = q.QVBoxLayout(self)
        layout.addWidget(label)
        layout.addWidget(self.tedit)
        btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Save | q.QDialogButtonBox.Cancel);
        layout.addWidget(btn_box)

        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)

    def accept(self) -> None:
        try:
            data = yaml.load(self.tedit.toPlainText(), Loader=yaml.SafeLoader)
        except yaml.parser.ParserError as e:
            q.QMessageBox.critical(self, 'Invalid data', f'Could not parse array data. Please verify that the syntax is correct and try again.\n\nDetails:\n\n{e}')
            return

        if not isinstance(data, list):
            q.QMessageBox.critical(self, 'Invalid data', 'Not an array')
            return
        if not data:
            q.QMessageBox.critical(self, 'Invalid data', f'The array must contain at least one element, and all elements should be {type(self.data[0]).__name__}s.')
            return
        if not util.are_list_types_homogeneous_and_equal(self.data, data):
            q.QMessageBox.critical(self, 'Invalid data', f'All elements should be {type(self.data[0]).__name__}s.')
            return

        self.data = data
        super().accept()

def _abstract_item_view_edit(parent, model, index: qc.QModelIndex, trigger, triggers) -> bool:
    if not (triggers & trigger) or not index.isValid() or not (index.flags() & (qc.Qt.ItemIsEditable | qc.Qt.ItemIsUserCheckable)):
        return False

    data = index.data(qc.Qt.UserRole)

    if isinstance(data, ActorIdentifier):
        dialog = ActorIdentifierEditDialog(parent, data)
        if dialog.exec_():
            model.setData(index, dialog.identifier, qc.Qt.EditRole)
        return True

    if isinstance(data, list):
        array_dialog = ArrayEditDialog(parent, data)
        if array_dialog.exec_():
            model.setData(index, array_dialog.data, qc.Qt.EditRole)
        return True

    return False

class CustomTableView(q.QTableView):
    def edit(self, index: qc.QModelIndex, trigger, event) -> bool:
        if _abstract_item_view_edit(self, self.model(), index, trigger, self.editTriggers()):
            return False
        return super().edit(index, trigger, event)



import typing

from evfl import EventFlow, Actor
from evfl.common import StringHolder
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class ActorStringListModel(qc.QAbstractListModel):
    def __init__(self, parent, l: typing.List[StringHolder]) -> None:
        super().__init__(parent)
        self.l = l

    def set(self, l: typing.List[StringHolder]) -> None:
        self.beginResetModel()
        self.l = l
        self.endResetModel()

    def rowCount(self, parent) -> int:
        return len(self.l)

    def flags(self, index: qc.QModelIndex) -> qc.Qt.ItemFlags:
        return qc.Qt.ItemIsEditable | super().flags(index)

    def has(self, string: str) -> bool:
        return any(s.v == string for s in self.l)

    def append(self, string: str) -> bool:
        self.beginInsertRows(qc.QModelIndex(), len(self.l), len(self.l))
        self.l.append(StringHolder(string))
        self.endInsertRows()
        return True

    def remove(self, row: int) -> bool:
        self.beginRemoveRows(qc.QModelIndex(), row, row)
        self.l.pop(row)
        self.endRemoveRows()
        return True

    def setData(self, index: qc.QModelIndex, value, role) -> bool:
        if role != qc.Qt.EditRole:
            return False
        row = index.row()
        assert isinstance(value, str)
        self.l[row].v = value
        self.dataChanged.emit(index, index)
        return True

    def data(self, index: qc.QModelIndex, role) -> qc.QVariant:
        row = index.row()
        if role == qc.Qt.UserRole:
            return self.l[row]
        if role == qc.Qt.DisplayRole or role == qc.Qt.ToolTipRole or role == qc.Qt.EditRole:
            return str(self.l[row])
        return qc.QVariant()



import copy
import typing

import eventeditor.ai as ai
import eventeditor.actor_json as aj
from eventeditor.actor_string_list_model import ActorStringListModel
from eventeditor.container_model import ContainerModel
from eventeditor.container_view import ContainerView
from eventeditor.flow_data import FlowData, FlowDataChangeReason
import eventeditor.util as util
from evfl import Container, Actor, Event
from evfl.enums import EventType
import evfl.event
import json
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtGui as qg # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class ActorProxyModel(qc.QIdentityProxyModel):
    def data(self, index, role):
        if index.column() == 0 and role == qc.Qt.DisplayRole:
            return str(self.sourceModel().data(index, qc.Qt.UserRole).identifier)
        return super().data(index, role)

class ActorRelatedEventEditDialog(q.QDialog):
    def __init__(self, parent, flow_data: FlowData, idx: int, attr_list_name: str, attr_name: str) -> None:
        super().__init__(parent, qc.Qt.WindowTitleHint | qc.Qt.WindowSystemMenuHint)
        self.setWindowTitle('Edit event')
        self.setMinimumWidth(700)
        self.setMinimumHeight(500)
        self.flow_data = flow_data
        self.event = self.flow_data.event_model.createIndex(idx, 0).data(qc.Qt.UserRole)
        assert isinstance(self.event.data, evfl.event.ActionEvent) or isinstance(self.event.data, evfl.event.SwitchEvent)
        self.is_switch = isinstance(self.event.data, evfl.event.SwitchEvent)
        self.attr_list_name = attr_list_name
        self.attr_name = attr_name

        if self.is_switch:
            self.setWindowTitle('Edit switch event')
        else:
            self.setWindowTitle('Edit action event')

        self.param_model = ContainerModel(self)
        if not self.event.data.params:
            self.event.data.params = Container()
        self.modified_params: Container = copy.deepcopy(self.event.data.params)
        self.param_model.set(self.modified_params)
        self.attr_model = ActorStringListModel(self, [])
        util.connect_model_change_signals(self.attr_model, self.flow_data, FlowDataChangeReason.Actors)

        self.createActorCbox()
        self.createAttrCbox()
        self.createParametersView()

        row = q.QHBoxLayout()
        row.addWidget(self.actor_cbox, stretch=1)
        separator = q.QLabel('::')
        row.addWidget(separator)
        row.addWidget(self.attr_cbox, stretch=1)

        layout = q.QVBoxLayout(self)
        layout.addLayout(row)
        layout.addWidget(self.param_view)
        btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Save | q.QDialogButtonBox.Cancel)
        layout.addWidget(btn_box)

        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)

    def createActorCbox(self) -> None:
        self.actor_proxy_model = ActorProxyModel(self)
        self.actor_proxy_model.setSourceModel(self.flow_data.actor_model)
        self.actor_cbox = q.QComboBox()
        self.actor_cbox.currentIndexChanged.connect(self.onActorSelected)
        self.actor_cbox.setModel(self.actor_proxy_model)
        actor = self.event.data.actor.v
        self.actor_cbox.setCurrentIndex(self.actor_cbox.findData(actor))

    def createAttrCbox(self) -> None:
        self.attr_cbox = q.QComboBox()
        self.attr_cbox.setModel(self.attr_model)
        attr = getattr(self.event.data, self.attr_name).v
        self.attr_cbox.setCurrentIndex(self.attr_cbox.findData(attr))

    def createParametersView(self) -> None:
        self.param_view = ContainerView(None, self.param_model, self.flow_data, has_autofill_btn=True)
        self.param_view.autofillRequested.connect(self.onAutofillRequested)
        self.param_view.reorderRequested.connect(self.onReorderRequested)
        self.param_view.copyJsonRequested.connect(self.onCopyJsonRequested)
        self.param_view.pasteJsonRequested.connect(self.onPasteJsonRequested)

    def onAutofillRequested(self) -> None:
        new_actor: Actor = self.actor_cbox.currentData()
        new_attr: str = self.attr_cbox.currentData().v if self.attr_cbox.currentData() else ''
        if not new_actor or not new_attr:
            q.QMessageBox.critical(self, 'Cannot auto fill', 'Please select an actor and a function.')
            return

        aiprog = ai.load_aiprog(new_actor.identifier.name)
        if not aiprog:
            if not self.tryJsonAutofill(new_actor.identifier.name, new_attr):
                q.QMessageBox.critical(self, 'Cannot auto fill', 'Failed to load the actor AI program')
            return

        actual_ai_class: typing.Optional[str] = None
        if self.is_switch:
            actual_ai_class = aiprog.queries.get(new_attr, None)
        else:
            actual_ai_class = aiprog.actions.get(new_attr, None)

        if actual_ai_class is None:
            q.QMessageBox.critical(self, 'Cannot auto fill', 'The selected action/query is not registered in the AI program.')
            return

        ai_type = ai.AIType.Query if self.is_switch else ai.AIType.Action
        parameters = ai.ai_def_instance.get_parameters(ai_type, actual_ai_class)

        self.modified_params.data.clear()
        if not self.is_switch:
            self.modified_params.data['IsWaitFinish'] = False
        for param in parameters:
            self.modified_params.data[param.name] = param.get_default_value()

        self.param_model.set(self.modified_params)
    
    def tryJsonAutofill(self, actor_name: str, attr_name: str) -> bool:        
        try:
            event_type = aj.EventType.Query if self.is_switch else aj.EventType.Action
            parameters = aj.load_event_parameters(actor_name, attr_name, event_type)

            if parameters is None:
                return False

            self.modified_params.data.clear()
            for param in parameters:
                self.modified_params.data[param] = parameters[param]
            self.param_model.set(self.modified_params)

            return True

        except:
            return False
    
    def onReorderRequested(self) -> None:
        new_actor: Actor = self.actor_cbox.currentData()
        new_attr: str = self.attr_cbox.currentData().v if self.attr_cbox.currentData() else ''
        if not new_actor or not new_attr:
            q.QMessageBox.critical(self, 'Cannot reorder parameters', 'Please select an actor and a function.')
            return

        event_type = aj.EventType.Query if self.is_switch else aj.EventType.Action
        parameters = aj.load_event_parameters(new_actor.identifier.name, new_attr, event_type)

        if parameters is None:
            q.QMessageBox.critical(self, 'Cannot reorder parameters', 'Failed to load the actor definition.')
            return

        # Parameters not contained in the definition will be left at the start of the collection
        for param in parameters:
            if param in self.modified_params.data:
                self.modified_params.data[param] = self.modified_params.data.pop(param)
        self.param_model.set(self.modified_params)

    def onCopyJsonRequested(self) -> None:
        toClipboard = json.dumps(self.modified_params.data)

        if self.attr_cbox.currentData():
            params = dict()
            params[str(self.attr_cbox.currentData().v)] = self.modified_params.data
            # Remove outer curly brackets for convenience
            toClipboard = json.dumps(params)[1:-1]

        q.QApplication.clipboard().setText(toClipboard)

    def onPasteJsonRequested(self) -> None:
        try:
            data = json.loads(f'{{{q.QApplication.clipboard().text()}}}')
            params = data[self.attr_cbox.currentData().v]
            self.modified_params.data.clear()
            for param in params:
                self.modified_params.data[param] = params[param]
            self.param_model.set(self.modified_params)

        except:
            q.QMessageBox.critical(self, 'Paste JSON', 'Failed to paste clipboard data as parameters.')

    def onActorSelected(self, actor_idx: int) -> None:
        if actor_idx == -1:
            return
        self.attr_model.set(getattr(self.actor_cbox.currentData(), self.attr_list_name))

    def accept(self) -> None:
        new_actor = self.actor_cbox.currentData()
        new_attr = self.attr_cbox.currentData()
        if not new_actor or not new_attr:
            q.QMessageBox.critical(self, 'Invalid data', 'Please select an actor and a function.')
            return

        previous_actor = self.event.data.actor.v
        previous_attr = getattr(self.event.data, self.attr_name).v
        self.event.data.actor.v = new_actor
        attr = getattr(self.event.data, self.attr_name)
        attr.v = new_attr

        reason = FlowDataChangeReason.Unknown
        if previous_actor != new_actor or previous_attr != new_attr:
            reason |= FlowDataChangeReason.Events
        if self.event.data.params.data != self.modified_params.data:
            reason |= FlowDataChangeReason.EventParameters

        self.event.data.params = self.modified_params

        self.flow_data.flowDataChanged.emit(reason)
        super().accept()

class SubFlowEventEditDialog(q.QDialog):
    def __init__(self, parent, flow_data: FlowData, idx: int) -> None:
        super().__init__(parent, qc.Qt.WindowTitleHint | qc.Qt.WindowSystemMenuHint)
        self.setWindowTitle('Edit event')
        self.setMinimumWidth(500)
        self.flow_data = flow_data
        self.event = self.flow_data.event_model.createIndex(idx, 0).data(qc.Qt.UserRole)
        assert self.flow_data.flow and isinstance(self.event.data, evfl.event.SubFlowEvent)
        self.param_model = ContainerModel(self)
        if not self.event.data.params:
            self.event.data.params = Container()
        self.modified_params = copy.deepcopy(self.event.data.params)
        self.param_model.set(self.modified_params)

        form = q.QFormLayout()

        self.flowchart_ledit = q.QLineEdit()
        self.flowchart_ledit.setText(self.event.data.res_flowchart_name)
        self.flowchart_ledit.setPlaceholderText('Flowchart name (optional)')
        form.addRow('&Flowchart:', self.flowchart_ledit)
        self.entry_point_ledit = q.QLineEdit()
        self.entry_point_ledit.setText(self.event.data.entry_point_name)
        self.entry_point_ledit.setPlaceholderText('Entry point name (mandatory)')
        form.addRow('&Entry point:', self.entry_point_ledit)

        help_msg = q.QLabel(f'Note: this flowchart ({self.flow_data.flow.name}) will be used if no flowchart is explicitly specified.')

        btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Save | q.QDialogButtonBox.Cancel);
        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)
        layout = q.QVBoxLayout(self)
        layout.addLayout(form)
        layout.addWidget(help_msg)
        layout.addWidget(ContainerView(None, self.param_model, self.flow_data))
        layout.addWidget(btn_box)

    def accept(self) -> None:
        new_flowchart = self.flowchart_ledit.text()
        new_ep = self.entry_point_ledit.text()
        if not new_ep:
            q.QMessageBox.critical(self, 'Invalid data', 'The entry point name cannot be empty.')
            return

        prev_flowchart = self.event.data.res_flowchart_name
        prev_ep = self.event.data.entry_point_name
        self.event.data.res_flowchart_name = new_flowchart
        self.event.data.entry_point_name = new_ep

        reason = FlowDataChangeReason.Unknown
        if prev_flowchart != new_flowchart or prev_ep != new_ep:
            reason |= FlowDataChangeReason.Events
        if self.event.data.params.data != self.modified_params.data:
            reason |= FlowDataChangeReason.EventParameters

        self.event.data.params = self.modified_params

        self.flow_data.flowDataChanged.emit(reason)
        super().accept()

def make_event_edit_dialog(parent, flow_data: FlowData, idx: int) -> typing.Optional[q.QDialog]:
    model = flow_data.event_model
    event = flow_data.event_model.data(model.createIndex(idx, 0), qc.Qt.UserRole)
    if isinstance(event.data, evfl.event.ActionEvent):
        return ActorRelatedEventEditDialog(parent, flow_data, idx, 'actions', 'actor_action')
    if isinstance(event.data, evfl.event.SwitchEvent):
        return ActorRelatedEventEditDialog(parent, flow_data, idx, 'queries', 'actor_query')
    if isinstance(event.data, evfl.event.SubFlowEvent):
        return SubFlowEventEditDialog(parent, flow_data, idx)
    return None

def show_event_editor(parent, flow_data: FlowData, idx: int) -> bool:
    dialog = make_event_edit_dialog(parent, flow_data, idx)
    if dialog:
        dialog.exec_()
        return True

    q.QMessageBox.information(parent, 'Edit event', 'This event has no editable property.')
    return False



import re
import typing

from evfl import EventFlow
from evfl.entry_point import EntryPoint
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class EntryPointModel(qc.QAbstractListModel):
    def __init__(self, *kwargs) -> None:
        super().__init__(*kwargs)
        self.flow: typing.Optional[EventFlow] = None
        self.l: typing.List[EntryPoint] = []

    def append(self, entry_point: EntryPoint) -> bool:
        self.beginInsertRows(qc.QModelIndex(), len(self.l), len(self.l))
        self.l.append(entry_point)
        self.endInsertRows()
        return True

    def removeRow(self, row: int) -> bool:
        self.beginRemoveRows(qc.QModelIndex(), row, row)
        self.l.pop(row)
        self.endRemoveRows()
        return True

    def flags(self, index: qc.QModelIndex) -> qc.Qt.ItemFlags:
        return qc.Qt.ItemIsEditable | super().flags(index)

    def has(self, name: str) -> bool:
        return any(entry.name == name for entry in self.l)

    def set(self, flow) -> None:
        self.beginResetModel()
        self.flow = flow
        self.l = self.flow.flowchart.entry_points if self.flow and self.flow.flowchart else []
        self.endResetModel()

    def rowCount(self, parent) -> int:
        return len(self.l)

    def setData(self, index: qc.QModelIndex, value, role) -> bool:
        if role != qc.Qt.EditRole or not index.isValid():
            return False
        if not isinstance(value, str) or not value or re.match('^Event(\d)+$', value) is not None or value == '*j32':
            q.QMessageBox.critical(None, 'Cannot rename', f'"{value}" is an invalid entry point name.')
            return False
        if self.l[index.row()].name == value:
            return False
        if self.has(value):
            q.QMessageBox.critical(None, 'Cannot rename', f'"{value}" is already used by another entry point.')
            return False
        self.l[index.row()].name = value
        self.dataChanged.emit(index, index)
        return True

    def data(self, index: qc.QModelIndex, role):
        if role == qc.Qt.UserRole:
            return self.l[index.row()]
        if role == qc.Qt.DisplayRole or role == qc.Qt.EditRole or role == qc.Qt.ToolTipRole:
            return self.l[index.row()].name
        return qc.QVariant()



from enum import IntEnum, auto
import typing
from typing import Union, Any
import threading
from PyQt5 import QtCore as qc, QtWidgets as q
from PyQt5.QtCore import QVariant
from evfl import EventFlow, Event
import chatgpt_events as cge
from eventeditor.util import get_event_type, get_event_description, get_event_next_summary, get_event_param_list

class EventModelColumn(IntEnum):
    Name = 0
    Type = auto()
    Description = auto()
    Next = auto()
    Parameters = auto()

class EventModel(qc.QAbstractTableModel):
    def __init__(self, *kwargs):
        super().__init__(*kwargs)
        self.flow: typing.Optional[EventFlow] = None
        self.l: list = []

    def append(self, event: Event) -> bool:
        self.beginInsertRows(qc.QModelIndex(), len(self.l), len(self.l))
        self.l.append(event)
        self.endInsertRows()
        return True

    def removeRow(self, row: int) -> bool:
        if row < 0 or row >= len(self.l):
            return False
        self.beginRemoveRows(qc.QModelIndex(), row, row)
        self.l.pop(row)
        self.endRemoveRows()
        return True

    def set(self, flow) -> None:
        self.beginResetModel()
        self.flow = flow
        self.l = self.flow.flowchart.events if self.flow and self.flow.flowchart else []
        self.endResetModel()

    def columnCount(self, parent=qc.QModelIndex()) -> int:
        return len(EventModelColumn)

    def rowCount(self, parent=qc.QModelIndex()) -> int:
        return len(self.l)

    def headerData(self, section, orientation, role) -> Union[QVariant, QVariant, str]:
        if role != qc.Qt.DisplayRole:
            return QVariant()

        column = EventModelColumn(section)
        return {
            EventModelColumn.Name: 'Name',
            EventModelColumn.Type: 'Type',
            EventModelColumn.Description: 'Description',
            EventModelColumn.Next: 'Next',
            EventModelColumn.Parameters: 'Parameters'
        }.get(column, 'Unknown')

    def data(self, index, role) -> Union[QVariant, str, Any]:
        if not index.isValid() or not (0 <= index.row() < len(self.l)):
            return QVariant()
        event = self.l[index.row()]
        if role in (qc.Qt.DisplayRole, qc.Qt.ToolTipRole):
            column = EventModelColumn(index.column())
            if column == EventModelColumn.Name:
                return event.name
            elif column == EventModelColumn.Type:
                return get_event_type(event)
            elif column == EventModelColumn.Description:
                return get_event_description(event)
            elif column == EventModelColumn.Next:
                return get_event_next_summary(event)
            elif column == EventModelColumn.Parameters:
                params = get_event_param_list(event)
                return '; '.join([f'{k}={v}' for k, v in params.items()]) if role == qc.Qt.DisplayRole else '<br>'.join(
                    [f'<b>{k}</b>: {v}' for k, v in params.items()])
        return QVariant()

    def generate_event_description(self, event_index):
        if not (0 <= event_index < len(self.l)):
            return
        event = self.l[event_index]
        prompt = f"Generate a description for an event where {event.name}"
        def handle_description(text):
            event.description = text
            idx = self.index(event_index, EventModelColumn.Description)
            self.dataChanged.emit(idx, idx)
        thread = threading.Thread(target=lambda: handle_description(cge.generate_text(prompt)))
        thread.start()



from enum import IntEnum, auto
import typing

from eventeditor.event_edit_dialog import show_event_editor
from eventeditor.event_model import EventModelColumn
from eventeditor.search_bar import SearchBar
from evfl import Event, EventFlow
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtGui as qg # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class _TableWidget(q.QTableView):
    onEnterPressed = qc.pyqtSignal()
    def keyPressEvent(self, event):
        key = event.key()
        if key == qc.Qt.Key_Return or key == qc.Qt.Key_Enter:
            self.onEnterPressed.emit()
        else:
            super().keyPressEvent(event)

class EventView(q.QWidget):
    jumpToFlowchartRequested = qc.pyqtSignal(int)

    def __init__(self, parent, flow_data, enable_ctx_menu: bool=True) -> None:
        super().__init__(parent)
        self.flow_data = flow_data
        self.enable_ctx_menu = enable_ctx_menu
        self.initWidgets()
        self.initLayout()
        self.connectWidgets()

    def initWidgets(self) -> None:
        self.event_proxy_model = qc.QSortFilterProxyModel(self)
        self.event_proxy_model.setSourceModel(self.flow_data.event_model)
        self.event_view = _TableWidget()
        self.event_view.setModel(self.event_proxy_model)
        self.event_view.verticalHeader().hide()
        self.event_view.setSelectionBehavior(q.QAbstractItemView.SelectRows)
        self.event_view.setSelectionMode(q.QAbstractItemView.SingleSelection)
        self.event_view.horizontalHeader().setMinimumSectionSize(80)
        self.event_view.horizontalHeader().setSectionResizeMode(q.QHeaderView.ResizeToContents)
        self.event_view.horizontalHeader().setSectionResizeMode(0, q.QHeaderView.ResizeToContents)
        self.event_view.horizontalHeader().setSectionResizeMode(2, q.QHeaderView.Stretch)
        self.event_view.horizontalHeader().setSectionResizeMode(3, q.QHeaderView.ResizeToContents)
        self.event_view.horizontalHeader().setSectionResizeMode(4, q.QHeaderView.Stretch)

        if self.enable_ctx_menu:
            self.event_view.setContextMenuPolicy(qc.Qt.CustomContextMenu)
            self.event_view.customContextMenuRequested.connect(self.onContextMenu)

        self.search_bar = SearchBar()

    def initLayout(self) -> None:
        layout = q.QVBoxLayout(self)
        layout.addWidget(self.event_view)
        layout.addWidget(self.search_bar)
        self.search_bar.hide()
        layout.setContentsMargins(0, 0, 0, 0)

    def connectWidgets(self) -> None:
        self.event_proxy_model.setFilterKeyColumn(-1)
        self.search_bar.connectToFilterModel(self.event_proxy_model)
        self.search_bar.addFindShortcut(self)

        self.event_view.onEnterPressed.connect(self.onEnterPressed)
        self.event_view.doubleClicked.connect(lambda idx: self.jumpToFlowchartRequested.emit(self.event_proxy_model.mapToSource(idx).row()))

    def getSelectedEventIdx(self) -> typing.Optional[qc.QModelIndex]:
        smodel = self.event_view.selectionModel()
        if not smodel.hasSelection():
            return None
        return self.event_proxy_model.mapToSource(smodel.selectedRows()[0])

    def getSelectedEvent(self) -> typing.Optional[Event]:
        source_idx = self.getSelectedEventIdx()
        if not source_idx:
            return None
        return source_idx.data(qc.Qt.UserRole)

    def selectEvent(self, event_idx: int) -> None:
        model = self.flow_data.event_model
        mapped_idx = self.event_proxy_model.mapFromSource(model.createIndex(event_idx, 0))
        if mapped_idx.isValid():
            self.event_view.selectRow(mapped_idx.row())

    def editEvent(self, event_idx: int) -> None:
        show_event_editor(self, self.flow_data, event_idx)

    def onContextMenu(self, pos) -> None:
        smodel = self.event_view.selectionModel()
        if not smodel.hasSelection():
            return

        sidx = smodel.selectedRows()[0]
        source_idx = self.event_proxy_model.mapToSource(sidx)
        menu = q.QMenu()
        menu.addAction('&Edit...', lambda: self.editEvent(source_idx.row()))
        menu.addAction('&Jump to flowchart', lambda: self.jumpToFlowchartRequested.emit(source_idx.row()))
        menu.exec_(self.sender().viewport().mapToGlobal(pos))

    def onEnterPressed(self) -> None:
        selected_idx = self.getSelectedEventIdx()
        if selected_idx:
            self.jumpToFlowchartRequested.emit(selected_idx.row())



# chatgpt_events.py
import google.generativeai as genai

# Configure the API key securely, possibly loaded from environment variables or a secure config
api_key = "AIzaSyCcICrkbWH8qV0VAZnRDzsEPNoO2_1U_Ps"
genai.configure(api_key=api_key)

def generate_text(prompt, model_name='gemini-pro'):
    """Generate text from a given prompt using a specified Gemini model."""
    model = genai.GenerativeModel(model_name)
    response = model.generate_content(prompt)
    return response.text

def generate_multimodal(prompt, image_bytes, model_name='gemini-pro-vision'):
    """Generate text using both text and image inputs."""
    model = genai.GenerativeModel(model_name)
    content = [prompt, image_bytes]  # Prepare the image_bytes appropriately
    response = model.generate_content(content)
    return response.text

def start_chat_conversation(initial_prompt, model_name='gemini-pro'):
    """Start a multi-turn conversation."""
    model = genai.GenerativeModel(model_name)
    chat = model.start_chat()
    response = chat.send_message(initial_prompt)
    return response.text, chat

def send_chat_message(chat, message):
    """Send a message in an ongoing chat conversation."""
    response = chat.send_message(message)
    return response.text



import typing

import eventeditor.actor_json as aj
from evfl import EventFlow
from evfl.event import ActionEvent, SwitchEvent

def reorder_event_flow_parameters(flow: EventFlow) -> None:
    # Should probably cache definition loads, though
    # current is useful for ensuring latest file version
    for event in flow.flowchart.events:
        if isinstance(event.data, ActionEvent):
            definition = aj.load_event_parameters(event.data.actor.v.identifier.name, event.data.actor_action.v.v, aj.EventType.Action)
        elif isinstance(event.data, SwitchEvent):
            definition = aj.load_event_parameters(event.data.actor.v.identifier.name, event.data.actor_query.v.v, aj.EventType.Query)
        else:
            continue

        # Nothing to sort
        if not event.data.params:
            continue

        # No defined order to sort by
        if not definition:
            continue

        reorder_event_parameters(event.data.params.data, definition)

# Parameters not contained in the definition will be left at the start of the collection
def reorder_event_parameters(params: typing.Dict[str, typing.Any], definition: typing.Iterable[str]) -> None:
    for param in definition:
        if param in params:
            params[param] = params.pop(param)



import enum
from eventeditor.actor_model import ActorModel
from eventeditor.autosave import AutoSaveSystem
from eventeditor.entry_point_model import EntryPointModel
from eventeditor.event_model import EventModel
import eventeditor.util as util
from evfl import EventFlow
import PyQt5.QtCore as qc # type: ignore
import re
import typing

class FlowDataChangeReason(enum.Flag):
    Unknown = enum.auto()
    Reset = enum.auto()
    Actors = enum.auto()
    Events = enum.auto()
    EventParameters = enum.auto()
    EventFlowRename = enum.auto()

class FlowData(qc.QObject):
    flowDataChanged = qc.pyqtSignal(FlowDataChangeReason)
    fileLoaded = qc.pyqtSignal(EventFlow)

    def __init__(self) -> None:
        super().__init__()

        self.auto_save = AutoSaveSystem()
        self.fileLoaded.connect(lambda: self.auto_save.reset())
        self.flowDataChanged.connect(lambda reason: self.auto_save.save(self.flow))

        self.flow: typing.Optional[EventFlow] = None

        self.actor_model = ActorModel()
        self.entry_point_model = EntryPointModel()
        self.event_model = EventModel()

        util.connect_model_change_signals(self.actor_model, self, FlowDataChangeReason.Actors)
        util.connect_model_change_signals(self.entry_point_model, self, FlowDataChangeReason.Events)
        util.connect_model_change_signals(self.event_model, self, FlowDataChangeReason.Events)

        self._next_event_idx = 0

    def setFlow(self, flow: typing.Optional[EventFlow]) -> None:
        self.flow = flow
        self.actor_model.set(flow)
        self.entry_point_model.set(flow)
        self.event_model.set(flow)
        self.flowDataChanged.emit(FlowDataChangeReason.Reset)
        self.fileLoaded.emit(flow)

        self._next_event_idx = self.computeNextEventIdx()

    def computeNextEventIdx(self) -> int:
        if not self.flow or not self.flow.flowchart:
            return -1
        pattern = re.compile(r'^Event(\d+)$')
        max_id = 0
        for event in self.flow.flowchart.events:
            match = pattern.match(event.name)
            if match:
                max_id = max(max_id, int(match[1]))
        return max_id + 1

    def generateEventName(self) -> str:
        name = f'Event{self._next_event_idx}'
        self._next_event_idx += 1
        return name



from enum import IntEnum, auto
import typing

from eventeditor.actor_model import ActorModelColumn
from eventeditor.actor_string_list_model import ActorStringListModel
from eventeditor.actor_string_list_view import ActorActionListView, ActorQueryListView
from eventeditor.container_model import ContainerModel
from eventeditor.container_view import ContainerView
from eventeditor.flow_data import FlowDataChangeReason
import eventeditor.util as util
from evfl import Container, EventFlow, Actor, ActorIdentifier
from evfl.entry_point import EntryPoint
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class ActorEditDialog(q.QDialog):
    def __init__(self, parent, flow_data, idx: int) -> None:
        super().__init__(parent, qc.Qt.WindowTitleHint | qc.Qt.WindowSystemMenuHint)
        self.setWindowTitle('Edit actor')
        self.setMinimumWidth(500)
        self.flow_data = flow_data
        self.mapper = q.QDataWidgetMapper(self)
        self.mapper.setSubmitPolicy(q.QDataWidgetMapper.ManualSubmit)
        self.mapper.setModel(flow_data.actor_model)

        self.initWidgets(idx)
        self.initLayout()
        self.connectWidgets()

        self.mapper.setCurrentIndex(idx)
        self.prev_identifier = ActorIdentifier(self.name_edit.text(), self.sub_name_edit.text())

    def initWidgets(self, idx: int) -> None:
        self.form = q.QFormLayout()

        self.name_edit = q.QLineEdit()
        self.form.addRow('&Name:', self.name_edit)
        self.mapper.addMapping(self.name_edit, ActorModelColumn.Name)

        self.sub_name_edit = q.QLineEdit()
        self.form.addRow('S&ub name:', self.sub_name_edit)
        self.mapper.addMapping(self.sub_name_edit, ActorModelColumn.SubName)

        self.arg_group = self.createArgumentGroup()
        model = self.mapper.model()
        ep = model.data(model.createIndex(idx, -1), qc.Qt.UserRole).argument_entry_point.v
        if ep:
            self.arg_group.setChecked(True)
            self.ep_cbox.setCurrentIndex(self.ep_cbox.model().l.index(ep))
        else:
            self.arg_group.setChecked(False)

        self.btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Save | q.QDialogButtonBox.Cancel);

    def createArgumentGroup(self) -> q.QGroupBox:
        form = q.QFormLayout()

        self.argument_name_edit = q.QLineEdit()
        form.addRow('Argument name:', self.argument_name_edit)
        self.mapper.addMapping(self.argument_name_edit, ActorModelColumn.ArgumentName)

        self.ep_cbox = q.QComboBox()
        form.addRow('Entry point:', self.ep_cbox)
        self.ep_cbox.setModel(self.flow_data.entry_point_model)
        self.mapper.addMapping(self.ep_cbox, ActorModelColumn.ArgumentEntryPoint, b'currentData')

        group = q.QGroupBox('Is an &argument')
        group.setCheckable(True)
        group.setLayout(form)
        return group

    def initLayout(self) -> None:
        layout = q.QVBoxLayout(self)
        layout.addLayout(self.form)
        layout.addWidget(self.arg_group)
        layout.addWidget(self.btn_box)

    def connectWidgets(self) -> None:
        self.btn_box.accepted.connect(self.accept)
        self.btn_box.rejected.connect(self.reject)

    def accept(self) -> None:
        model = self.mapper.model()
        identifier = ActorIdentifier(self.name_edit.text(), self.sub_name_edit.text())
        if identifier != self.prev_identifier and model.has(identifier):
            q.QMessageBox.critical(self, 'Cannot edit actor', f'{identifier} is already used as an actor identifier. Please pick another one.')
            return

        self.mapper.submit()
        if not self.arg_group.isChecked():
            model.setData(model.createIndex(self.mapper.currentIndex(), ActorModelColumn.ArgumentName), '', qc.Qt.EditRole)
            model.setData(model.createIndex(self.mapper.currentIndex(), ActorModelColumn.ArgumentEntryPoint), None, qc.Qt.EditRole)
        super().accept()

class ActorAddDialog(ActorEditDialog):
    def __init__(self, parent, flow_data, idx) -> None:
        super().__init__(parent, flow_data, idx)
        self.setWindowTitle('Add new actor')

    def reject(self) -> None:
        self.flow_data.actor_model.remove(self.flow_data.actor_model.data(
            self.flow_data.actor_model.createIndex(self.flow_data.actor_model.rowCount(None) - 1, -1),
            qc.Qt.UserRole))
        super().reject()

class ActorDetailPane(q.QWidget):
    jumpToEventsRequested = qc.pyqtSignal(str)

    def __init__(self, parent, flow_data) -> None:
        super().__init__(parent)
        self.flow_data = flow_data
        self.actor: typing.Optional[Actor] = None
        self.action_model = ActorStringListModel(parent, [])
        self.query_model = ActorStringListModel(parent, [])
        self.container_model = ContainerModel(parent)
        self.initWidgets()
        self.initLayout()
        self.connectWidgets()

    def setActor(self, actor: typing.Optional[Actor]) -> None:
        self.actor = actor
        self.action_model.set(self.actor.actions if self.actor else [])
        self.query_model.set(self.actor.queries if self.actor else [])
        if self.actor and not self.actor.params:
            self.actor.params = Container()
        self.container_model.set(self.actor.params if self.actor else None)
        if self.actor:
            self.action_view.setActor(self.actor)
            self.query_view.setActor(self.actor)

    def initWidgets(self) -> None:
        self.action_view = ActorActionListView(self, self.action_model, self.flow_data)
        self.query_view = ActorQueryListView(self, self.query_model, self.flow_data)
        self.container_view = ContainerView(self, self.container_model, self.flow_data)

    def initLayout(self) -> None:
        layout = q.QHBoxLayout(self)

        layout.addWidget(self.action_view, stretch=1)
        layout.addWidget(self.query_view, stretch=1)
        layout.addWidget(self.container_view, stretch=2)
        layout.setContentsMargins(0, 0, 0, 0)

    def connectWidgets(self) -> None:
        util.connect_model_change_signals(self.action_model, self.flow_data, FlowDataChangeReason.Actors)
        util.connect_model_change_signals(self.query_model, self.flow_data, FlowDataChangeReason.Actors)
        util.connect_model_change_signals(self.container_model, self.flow_data, FlowDataChangeReason.Actors)

        self.action_view.addActionBuilder(lambda menu, idx: menu.addAction('&Jump to events', lambda: self.onJumpToEvents(idx)))
        self.query_view.addActionBuilder(lambda menu, idx: menu.addAction('&Jump to events', lambda: self.onJumpToEvents(idx)))
        self.container_view.addActionBuilder(lambda menu, idx: menu.addAction('&Add default create parameters', lambda: self.addDefaultCreateParameters()))

    def addDefaultCreateParameters(self) -> None:
        if not self.actor or not self.actor.params:
            return
        self.actor.params.data.update({'CreateMode': 0, 'IsGrounding': False, 'IsWorld': False, 'PosX': 0.0, 'PosY': 0.0, 'PosZ': 0.0, 'RotX': 0.0, 'RotY': 0.0, 'RotZ': 0.0})
        self.container_model.set(self.actor.params)

    def onJumpToEvents(self, idx) -> None:
        if self.actor:
            self.jumpToEventsRequested.emit(f'{self.actor.identifier}::{idx.data(qc.Qt.UserRole)}')

class ActorView(q.QWidget):
    jumpToActorEventsRequested = qc.pyqtSignal(str)

    def __init__(self, parent, flow_data) -> None:
        super().__init__(parent)
        self.flow_data = flow_data
        self.initWidgets()
        self.initLayout()
        self.connectWidgets()

    def initWidgets(self) -> None:
        self.actor_view = q.QTableView()
        self.actor_view.setModel(self.flow_data.actor_model)
        self.actor_view.verticalHeader().hide()
        self.actor_view.setSelectionBehavior(q.QAbstractItemView.SelectRows)
        self.actor_view.setSelectionMode(q.QAbstractItemView.SingleSelection)
        self.actor_view.horizontalHeader().setSectionResizeMode(q.QHeaderView.ResizeToContents)
        self.actor_view.horizontalHeader().setSectionResizeMode(0, q.QHeaderView.Stretch)
        self.actor_view.horizontalHeader().setSectionResizeMode(1, q.QHeaderView.Stretch)
        self.actor_view.setEditTriggers(q.QAbstractItemView.NoEditTriggers);
        self.actor_view.setContextMenuPolicy(qc.Qt.CustomContextMenu)
        self.actor_view.customContextMenuRequested.connect(self.onContextMenu)

        self.detail_pane = ActorDetailPane(None, self.flow_data)

        self.top_box = q.QHBoxLayout()
        self.top_box.setContentsMargins(5,5,5,0)
        self.num_actors_label = q.QLabel()
        self.top_box.addWidget(self.num_actors_label, stretch=1)
        self.add_actor_btn = q.QPushButton('A&dd...')
        self.add_actor_btn.setEnabled(False)
        self.top_box.addWidget(self.add_actor_btn)

    def initLayout(self) -> None:
        splitter = q.QSplitter(qc.Qt.Vertical)
        splitter.addWidget(self.actor_view)
        self.stacked_pane = q.QStackedWidget()
        self.stacked_pane.addWidget(q.QWidget())
        self.stacked_pane.addWidget(self.detail_pane)
        splitter.addWidget(self.stacked_pane)
        splitter.setSizes([int(splitter.height() * 0.6), int(splitter.height() * 0.4)])
        layout = q.QVBoxLayout(self)
        layout.addLayout(self.top_box)
        layout.addWidget(splitter, stretch=1)
        layout.setContentsMargins(0, 0, 0, 0)

    def connectWidgets(self) -> None:
        self.add_actor_btn.clicked.connect(self.addActor)
        self.flow_data.fileLoaded.connect(lambda: self.add_actor_btn.setEnabled(True))
        self.flow_data.fileLoaded.connect(lambda: self.hideActorDetailPane())
        self.flow_data.flowDataChanged.connect(self.updateNumActorLabel)
        self.updateNumActorLabel()

        self.actor_view.doubleClicked.connect(self.editActor)
        self.actor_view.selectionModel().currentRowChanged.connect(self.onCurrentChanged)
        self.actor_view.selectionModel().selectionChanged.connect(self.onSelectionChanged)

    def updateNumActorLabel(self, *_) -> None:
        self.num_actors_label.setText(f'{self.flow_data.actor_model.rowCount(None)} actor(s)')

    def addActor(self) -> None:
        ok = self.flow_data.actor_model.appendEmptyActor()
        assert ok
        dialog = ActorAddDialog(self, self.flow_data, self.flow_data.actor_model.rowCount(None) - 1)
        dialog.exec_()

    def editActor(self, idx: qc.QModelIndex) -> None:
        dialog = ActorEditDialog(self, self.flow_data, idx.row())
        dialog.exec_()

    def removeActor(self, idx: qc.QModelIndex) -> None:
        actor = idx.data(qc.Qt.UserRole)
        if util.is_actor_in_use(self.flow_data.flow.flowchart.events, actor):
            q.QMessageBox.critical(self, 'Cannot remove actor', f'{actor.identifier} cannot be removed because it is used by events. Please remove any references to this actor and try again.')
            return

        self.flow_data.actor_model.remove(actor)

    def hideActorDetailPane(self) -> None:
        self.detail_pane.setActor(None)
        self.stacked_pane.setCurrentIndex(0)

    def onCurrentChanged(self, current, previous) -> None:
        if previous.row() == -1:
            self.hideActorDetailPane()

    def onSelectionChanged(self, selected, deselected) -> None:
        if len(selected.indexes()) != len(ActorModelColumn):
            self.hideActorDetailPane()
            return

        self.detail_pane.setActor(selected.indexes()[0].data(qc.Qt.UserRole))
        self.stacked_pane.setCurrentIndex(1)

    def onContextMenu(self, pos) -> None:
        smodel = self.actor_view.selectionModel()
        if not smodel.hasSelection():
            return

        idx = smodel.selectedRows()[0]

        menu = q.QMenu()
        menu.addAction('&Edit...', lambda: self.editActor(idx))
        menu.addAction('&Remove', lambda: self.removeActor(idx))
        menu.addAction('&Jump to events', lambda: self.jumpToActorEventsRequested.emit(str(idx.data(qc.Qt.UserRole).identifier) + '::'))
        menu.exec_(self.sender().viewport().mapToGlobal(pos))



import typing

import eventeditor.ai as ai
import eventeditor.actor_json as aj
import eventeditor.util as util
from eventeditor.search_bar import SearchBar
from evfl import EventFlow, Actor
from evfl.common import StringHolder
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class ActorStringListView(q.QWidget):
    def __init__(self, parent, label_str: str, model, flow_data) -> None:
        super().__init__(parent)
        self.flow_data = flow_data
        self.action_builders = [] # type: ignore
        self.model = model
        self.label_str = label_str

        self.lview = q.QListView()
        self.lview.setModel(self.model)
        self.lview.setContextMenuPolicy(qc.Qt.CustomContextMenu)
        self.lview.customContextMenuRequested.connect(self.onContextMenu)

        self.add_btn = q.QPushButton('Add...')
        self.add_btn.setStyleSheet('padding: 2px 5px;')
        self.add_btn.clicked.connect(self.onAdd)
        box = q.QHBoxLayout()
        label = q.QLabel(label_str)
        label.setStyleSheet('font-weight: bold;')
        box.addWidget(label, stretch=1)
        box.addWidget(self.add_btn)

        layout = q.QVBoxLayout(self)
        layout.addLayout(box)
        layout.addWidget(self.lview, stretch=1)

    def onAdd(self) -> None:
        text = self._getNewString()
        if not text:
            return

        if self.model.has(text):
            q.QMessageBox.critical(self, 'Cannot add', 'This action or query already exists.')
            return

        self.model.append(text)
        self.flow_data.actor_model.refresh()

    def _getNewString(self) -> str:
        text, ok = q.QInputDialog.getText(self, f'{self.label_str}', f'Name of the new action or query:', q.QLineEdit.Normal)
        return text

    def onRemove(self, idx) -> None:
        value = idx.data(qc.Qt.UserRole)
        if util.is_actor_string_in_use(self.flow_data.flow.flowchart.events, value):
            q.QMessageBox.critical(self, 'Cannot remove', 'This action or query cannot be removed because it is in use. Please remove any references to this action or query first.')
            return
        self.model.remove(idx.row())
        self.flow_data.actor_model.refresh()

    def addActionBuilder(self, fn) -> None:
        self.action_builders.append(fn)

    def onContextMenu(self, pos) -> None:
        smodel = self.lview.selectionModel()
        if not smodel.selectedRows():
            return

        idx = smodel.selectedRows()[0]
        menu = q.QMenu()
        menu.addAction('&Remove', lambda: self.onRemove(idx))
        for builder in self.action_builders:
            builder(menu, idx)
        menu.exec_(self.sender().viewport().mapToGlobal(pos))

class ActorAIClassAddDialog(q.QDialog):
    def __init__(self, parent, model) -> None:
        super().__init__(parent, qc.Qt.WindowTitleHint | qc.Qt.WindowSystemMenuHint)
        self.setWindowTitle('Add an AI class')
        self.setMinimumWidth(350)

        ledit_hint = q.QLabel('Enter an AI class:')
        ledit_hint.setAlignment(qc.Qt.AlignCenter)
        self._ledit = q.QLineEdit()
        list_hint = q.QLabel('or select one:')
        list_hint.setAlignment(qc.Qt.AlignCenter)

        self._list = q.QListView()
        self._proxy_model = qc.QSortFilterProxyModel(self)
        self._proxy_model.setSourceModel(model)
        self._list.setModel(self._proxy_model)
        self._list.setEditTriggers(q.QAbstractItemView.NoEditTriggers)
        self._list.selectionModel().selectionChanged.connect(self._onSelectionChanged)
        self._list.doubleClicked.connect(lambda idx: self.accept())

        self._search_bar = SearchBar()
        self._search_bar.hide()
        self._search_bar.connectToFilterModel(self._proxy_model)
        self._search_bar.addFindShortcut(self)

        btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Ok | q.QDialogButtonBox.Cancel);
        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)
        layout = q.QVBoxLayout(self)
        layout.addWidget(ledit_hint)
        layout.addWidget(self._ledit)
        layout.addWidget(list_hint)
        layout.addWidget(self._list)
        layout.addWidget(self._search_bar)
        layout.addWidget(btn_box)

    def accept(self) -> None:
        if not self.getText():
            q.QMessageBox.critical(self, self.windowTitle(), 'Please enter or select an AI class.')
            return
        super().accept()

    def getText(self) -> str:
        return self._ledit.text()

    def _onSelectionChanged(self, selected, deselected) -> None:
        if len(selected.indexes()) <= 0:
            return
        self._ledit.setText(selected.indexes()[0].data(qc.Qt.DisplayRole))

class ActorActionListView(ActorStringListView):
    def __init__(self, parent, model, flow_data) -> None:
        super().__init__(parent, 'Actions', model, flow_data)
        self.actor: typing.Optional[Actor] = None

    def setActor(self, actor: Actor) -> None:
        self.actor = actor

    def _getNewString(self) -> str:
        if not self.actor:
            return ''
        name = self.actor.identifier.name

        actions = []
        aiprog = ai.load_aiprog(name)
        if aiprog:
            actions = list(aiprog.actions.keys())
        else:
            json_actions = aj.load_actions(name)
            if json_actions:
                actions = list(json_actions)

        add_dialog = ActorAIClassAddDialog(self, qc.QStringListModel(actions, self))
        add_dialog.setWindowTitle(f'Add an action for {name}')
        ret = add_dialog.exec_()
        return add_dialog.getText() if ret else ''

class ActorQueryListView(ActorStringListView):
    def __init__(self, parent, model, flow_data) -> None:
        super().__init__(parent, 'Queries', model, flow_data)
        self.actor: typing.Optional[Actor] = None

    def setActor(self, actor: Actor) -> None:
        self.actor = actor

    def _getNewString(self) -> str:
        if not self.actor:
            return ''
        name = self.actor.identifier.name

        queries = []
        aiprog = ai.load_aiprog(name)
        if aiprog:
            queries = list(aiprog.queries.keys())
        else:
            json_queries = aj.load_queries(name)
            if json_queries:
                queries = list(json_queries)

        add_dialog = ActorAIClassAddDialog(self, qc.QStringListModel(queries, self))
        add_dialog.setWindowTitle(f'Add a query for {name}')
        ret = add_dialog.exec_()
        return add_dialog.getText() if ret else ''



import eventeditor.util as util
from evfl import EventFlow
from pathlib import Path
import PyQt5.QtCore as qc # type: ignore
import queue
import sys
import threading
import traceback
import typing

class TaskQueue(queue.Queue):
    def __init__(self):
        super().__init__()
        t = threading.Thread(target=self._thread_func)
        t.daemon = True
        t.start()

    def add_task(self, task):
        self.put(task)

    def _thread_func(self):
        while True:
            item = self.get()
            item()
            self.task_done()

class AutoSaveSystem:
    def __init__(self) -> None:
        self._save_dir: typing.Optional[Path] = None
        save_dir = qc.QStandardPaths.writableLocation(qc.QStandardPaths.AppLocalDataLocation)
        if not save_dir:
            return
        self._save_dir = Path(save_dir)
        self._save_dir.mkdir(parents=True, exist_ok=True)
        self._queue = TaskQueue()
        self.reset()

    def get_directory(self) -> typing.Optional[Path]:
        return self._save_dir

    def reset(self) -> None:
        if not self._save_dir:
            return

        self._queue.join()
        self._current_save_idx = 0

    def save(self, flow: typing.Optional[EventFlow]) -> None:
        if not self._save_dir:
            return

        def do_write():
            if not flow or not self._save_dir:
                return
            path = self._save_dir/f'autosave_{flow.name}__{self._current_save_idx}.bfevfl.gz'
            try:
                util.write_flow(str(path), flow)
                self._current_save_idx = (self._current_save_idx + 1) % 10
            except:
                sys.stderr.write(f'!!! Autosave failed !!!\n{traceback.format_exc()}\n\n')

        self._queue.add_task(do_write)



from enum import IntEnum, auto
import typing
import yaml

from eventeditor.container_model import ContainerModel, ContainerModelColumn
from eventeditor.data_editors import CustomTableView
import eventeditor.util as util
from evfl import ActorIdentifier, Argument, Container
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtGui as qg # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class ContainerAddItemDialogType(IntEnum):
    Value = 0
    ActorIdentifier = auto()
    Argument = auto()

class ContainerAddItemDialog(q.QDialog):
    def __init__(self, parent, model: ContainerModel) -> None:
        super().__init__(parent, qc.Qt.WindowTitleHint | qc.Qt.WindowSystemMenuHint)
        self.setWindowTitle(f'Add container item')
        self.resize(500, 100)
        self.model = model

        self.key_ledit = q.QLineEdit()
        self.createTypeWidgets()
        self.createValueWidgets()

        form_layout = q.QFormLayout()
        form_layout.addRow('&Key:', self.key_ledit)
        form_layout.addRow('Type:', self.type_layout)
        form_layout.addRow('&Value:', self.value_widget)

        layout = q.QVBoxLayout(self)
        layout.addLayout(form_layout)
        btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Save | q.QDialogButtonBox.Cancel)
        layout.addWidget(btn_box)

        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)

    def createTypeWidgets(self) -> None:
        self.type_group = q.QButtonGroup()
        self.type_layout = q.QHBoxLayout()
        def add_type_choice(label_text, btn_id):
            rbtn = q.QRadioButton(label_text)
            self.type_group.addButton(rbtn, btn_id)
            self.type_layout.addWidget(rbtn)
        add_type_choice('&Int/float/bool/string/array', ContainerAddItemDialogType.Value)
        add_type_choice('&Actor identifier', ContainerAddItemDialogType.ActorIdentifier)
        add_type_choice('Ar&gument', ContainerAddItemDialogType.Argument)
        self.type_group.button(ContainerAddItemDialogType.Value).setChecked(True)
        self.type_layout.addStretch()
        self.type_group.buttonClicked[int].connect(self.onTypeBtnClicked)

    def createValueWidgets(self) -> None:
        self.value_widget = q.QStackedWidget()

        self.tedit = q.QPlainTextEdit()
        font = qg.QFontDatabase.systemFont(qg.QFontDatabase.FixedFont)
        font.setPointSize(int(font.pointSize() * 1.05))
        self.tedit.setFont(font)
        self.tedit.setPlaceholderText('Item data (YAML)\nExamples:\n[1, 2, 3]\n3.1415\ntrue')
        self.value_widget.addWidget(self.tedit)

        actor_id_widget = q.QWidget()
        actor_id_form = q.QFormLayout(actor_id_widget)
        self.actor_id_name_box = q.QLineEdit()
        actor_id_form.addRow('&Name:', self.actor_id_name_box)
        self.actor_id_sub_name_box = q.QLineEdit()
        actor_id_form.addRow('&Sub name:', self.actor_id_sub_name_box)
        self.value_widget.addWidget(actor_id_widget)

        argument_widget = q.QWidget()
        argument_layout = q.QVBoxLayout(argument_widget)
        self.argument_box = q.QLineEdit()
        argument_layout.addWidget(self.argument_box)
        argument_layout.addStretch()
        self.value_widget.addWidget(argument_widget)

    def onTypeBtnClicked(self, btn_id: int) -> None:
        self.value_widget.setCurrentIndex(btn_id)

    def parseValue(self) -> typing.Any:
        try:
            data = yaml.load(self.tedit.toPlainText(), Loader=yaml.SafeLoader)
        except yaml.parser.ParserError as e:
            q.QMessageBox.critical(self, 'Invalid data', f'Could not parse item data. Please verify that the syntax is correct and try again.\n\nDetails:\n\n{e}')
            return None

        if isinstance(data, list):
            if not data:
                q.QMessageBox.critical(self, 'Invalid data', 'Arrays must contain at least one element.')
                return None
            if not util.is_valid_container_value_type(data[0]):
                q.QMessageBox.critical(self, 'Invalid data', f'{type(data[0]).__name__} is not a valid data type.')
                return None
            if not util.is_list_homogeneous(data):
                q.QMessageBox.critical(self, 'Invalid data', 'Arrays can only contain one element type.')
                return None
        else:
            if not util.is_valid_container_value_type(data):
                q.QMessageBox.critical(self, 'Invalid data', f'{type(data).__name__} is not a valid data type.')
                return None

        return data

    def parseActorIdentifier(self) -> typing.Any:
        identifier = ActorIdentifier(self.actor_id_name_box.text(), self.actor_id_sub_name_box.text())
        if not identifier.name:
            q.QMessageBox.critical(self, 'Invalid actor identifier', 'The actor name cannot be empty.')
            return None
        return identifier

    def parseArgument(self) -> typing.Any:
        argument = Argument(self.argument_box.text())
        if not argument:
            q.QMessageBox.critical(self, 'Invalid argument', 'The argument name cannot be empty.')
            return None
        return argument

    def accept(self) -> None:
        key_name: str = self.key_ledit.text()
        if not key_name:
            q.QMessageBox.critical(self, 'Invalid key', 'The key name cannot be empty.')
            return
        if self.model.has(key_name):
            q.QMessageBox.critical(self, 'Invalid key', f'{key_name} is already used.')
            return

        type_id = self.type_group.checkedId()
        data = None
        if type_id == ContainerAddItemDialogType.Value:
            data = self.parseValue()
        elif type_id == ContainerAddItemDialogType.ActorIdentifier:
            data = self.parseActorIdentifier()
        elif type_id == ContainerAddItemDialogType.Argument:
            data = self.parseArgument()

        if data is None:
            return

        self.model.insertItem(key_name, data)
        super().accept()

class ContainerView(q.QWidget):
    autofillRequested = qc.pyqtSignal()
    reorderRequested = qc.pyqtSignal()
    copyJsonRequested = qc.pyqtSignal()
    pasteJsonRequested = qc.pyqtSignal()

    def __init__(self, parent, model: ContainerModel, flow_data, has_autofill_btn=False) -> None:
        super().__init__(parent)
        self.flow_data = flow_data
        self.action_builders = [] # type: ignore
        self.model: ContainerModel = model

        self.tview = CustomTableView()
        font = qg.QFont()
        font.setPointSize(max(int(font.pointSize() * 0.85), 9))
        self.tview.setFont(font)
        self.tview.setModel(self.model)
        self.tview.verticalHeader().hide()
        self.tview.verticalHeader().setSectionResizeMode(q.QHeaderView.ResizeToContents)
        self.tview.setSelectionMode(q.QAbstractItemView.SingleSelection)
        hheader = self.tview.horizontalHeader()
        hheader.setSectionResizeMode(q.QHeaderView.ResizeToContents)
        hheader.setSectionResizeMode(ContainerModelColumn.Value, q.QHeaderView.Stretch)
        hheader.setMinimumSectionSize(50)
        self.tview.setContextMenuPolicy(qc.Qt.CustomContextMenu)
        self.tview.customContextMenuRequested.connect(self.onContextMenu)
        util.set_view_delegate(self.tview)

        self.add_btn = q.QPushButton('Add...')
        self.add_btn.setStyleSheet('padding: 2px 5px;')
        self.add_btn.clicked.connect(self.onAdd)
        self.autofill_btn = q.QPushButton('Auto fill')
        self.autofill_btn.setStyleSheet('padding: 2px 5px;')
        self.autofill_btn.clicked.connect(self.autofillRequested)
        self.reorder_btn = q.QPushButton('Reorder')
        self.reorder_btn.setStyleSheet('padding: 2px 5px;')
        self.reorder_btn.clicked.connect(self.reorderRequested)
        self.copy_btn = q.QPushButton('Copy JSON')
        self.copy_btn.setStyleSheet('padding: 2px 5px;')
        self.copy_btn.clicked.connect(self.copyJsonRequested)
        self.paste_btn = q.QPushButton('Paste JSON')
        self.paste_btn.setStyleSheet('padding: 2px 5px;')
        self.paste_btn.clicked.connect(self.pasteJsonRequested)
        box = q.QHBoxLayout()
        label = q.QLabel('Parameters')
        label.setStyleSheet('font-weight: bold;')
        box.addWidget(label, stretch=1)
        if has_autofill_btn:
            box.addWidget(self.copy_btn)
            box.addWidget(self.paste_btn)
            box.addWidget(self.reorder_btn)
            box.addWidget(self.autofill_btn)
        box.addWidget(self.add_btn)

        layout = q.QVBoxLayout(self)
        layout.addLayout(box)
        layout.addWidget(self.tview, stretch=1)

    def onAdd(self) -> None:
        dialog = ContainerAddItemDialog(self, self.model)
        dialog.exec_()

    def onRemove(self, idx) -> None:
        self.model.removeRow(idx.row())

    def onConvertToArgument(self, idx) -> None:
        self.model.changeTypeToArgument(idx.row())

    def addActionBuilder(self, fn) -> None:
        self.action_builders.append(fn)

    def onContextMenu(self, pos) -> None:
        smodel = self.tview.selectionModel()
        if not smodel.selectedIndexes():
            return
        idx = smodel.selectedIndexes()[0]
        menu = q.QMenu()
        menu.addAction('Convert to &argument', lambda: self.onConvertToArgument(idx))
        menu.addAction('&Remove item', lambda: self.onRemove(idx))
        for builder in self.action_builders:
            builder(menu, idx)
        menu.exec_(self.sender().viewport().mapToGlobal(pos))



import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtGui as qg # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class SearchBar(q.QWidget):
    textChanged = qc.pyqtSignal(str)
    caseInsensitiveChanged = qc.pyqtSignal(bool)

    def __init__(self, *args) -> None:
        super().__init__(*args)

        self.close_btn = q.QPushButton()
        self.close_btn.setIcon(self.style().standardIcon(q.QStyle.SP_DialogCloseButton))

        self.box = q.QLineEdit()
        self.box.setPlaceholderText('Search…')

        self.caseInsensitiveCbox = q.QCheckBox('Case insensitive')
        self.caseInsensitiveCbox.setChecked(True)

        layout = q.QHBoxLayout(self)
        layout.addWidget(self.close_btn)
        layout.addWidget(self.box)
        layout.addWidget(self.caseInsensitiveCbox)
        layout.setStretch(1, 1)
        layout.setContentsMargins(5, 5, 5, 5)

        self.close_btn.clicked.connect(self.hideAndClear)
        close_action = q.QAction(self)
        close_action.setShortcut(qg.QKeySequence.Cancel)
        close_action.triggered.connect(self.hideAndClear)
        self.addAction(close_action)

        self.box.textChanged.connect(self.textChanged)

        self.caseInsensitiveCbox.stateChanged.connect(lambda state: self.caseInsensitiveChanged.emit(state == qc.Qt.Checked))

    def hideAndClear(self) -> None:
        self.close()
        self.box.clear()

    def showAndFocus(self) -> None:
        self.show()
        self.box.setFocus()

    def setValue(self, value: str) -> None:
        self.box.setText(value)

    def addFindShortcut(self, widget) -> None:
        find_action = q.QAction(widget)
        find_action.setShortcut(qg.QKeySequence.Find)
        find_action.triggered.connect(self.showAndFocus)
        widget.addAction(find_action)

    def connectToFilterModel(self, model) -> None:
        self.textChanged.connect(model.setFilterFixedString)
        def setModelFilterCaseSensitivity(insensitive: bool) -> None:
            model.setFilterCaseSensitivity(qc.Qt.CaseInsensitive if insensitive else qc.Qt.CaseSensitive)
        self.caseInsensitiveChanged.connect(setModelFilterCaseSensitivity)
        setModelFilterCaseSensitivity(self.caseInsensitiveCbox.isChecked())



from enum import IntEnum, auto
import typing

import eventeditor.util as util
from evfl import ActorIdentifier, Argument, Container
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtWidgets as q # type: ignore

class ContainerModelColumn(IntEnum):
    DataType = 0
    Key = auto()
    Value = auto()

class ContainerModel(qc.QAbstractTableModel):
    def __init__(self, parent, container: typing.Optional[Container] = None) -> None:
        super().__init__(parent)
        self.set(container)

    def set(self, container: typing.Optional[Container]) -> None:
        self.beginResetModel()
        if container:
            self.cdata = container.data
            self.keys = list(container.data.keys())
        else:
            self.cdata = dict()
            self.keys = []
        self.endResetModel()

    def has(self, key: str) -> bool:
        return any(k == key for k in self.cdata.keys())

    def insertItem(self, key: str, value) -> bool:
        self.beginInsertRows(qc.QModelIndex(), len(self.keys), len(self.keys))
        self.cdata[key] = value
        self.keys.append(key)
        self.endInsertRows()
        return True

    def removeRow(self, row: int) -> bool:
        self.beginRemoveRows(qc.QModelIndex(), row, row)
        del self.cdata[self.keys[row]]
        self.keys.pop(row)
        self.endRemoveRows()
        return True

    def changeTypeToArgument(self, row: int) -> None:
        self.cdata[self.keys[row]] = Argument()
        self.dataChanged.emit(self.createIndex(row, ContainerModelColumn.DataType), self.createIndex(row, ContainerModelColumn.Value))

    def columnCount(self, parent) -> int:
        return len(ContainerModelColumn)

    def rowCount(self, parent) -> int:
        return len(self.cdata)

    def flags(self, index: qc.QModelIndex):
        if index.column() != ContainerModelColumn.Value:
            return super().flags(index)
        row = index.row()
        if isinstance(self.cdata[self.keys[row]], bool):
            return super().flags(index) | qc.Qt.ItemIsUserCheckable
        return super().flags(index) | qc.Qt.ItemIsEditable

    def headerData(self, section, orientation, role) -> qc.QVariant:
        if role != qc.Qt.DisplayRole:
            return qc.QVariant()
        if section == ContainerModelColumn.Key:
            return 'Key'
        if section == ContainerModelColumn.Value:
            return 'Value'
        if section == ContainerModelColumn.DataType:
            return 'Type'
        return 'Unknown'

    def setData(self, index: qc.QModelIndex, value, role: qc.Qt.ItemDataRole) -> bool:
        col = index.column()
        row = index.row()
        current_value = self.cdata[self.keys[row]]

        if role == qc.Qt.CheckStateRole:
            if col != ContainerModelColumn.Value or not isinstance(current_value, bool):
                return False
            self.cdata[self.keys[row]] = True if value == qc.Qt.Checked else False
            self.dataChanged.emit(index, index)
            return True

        if role != qc.Qt.EditRole:
            return False

        if isinstance(current_value, Argument):
            value = Argument(value)

        if isinstance(current_value, list) and not util.are_list_types_homogeneous_and_equal(value, current_value):
            q.QMessageBox.critical(None, 'Bug',
                f'Refusing to set value because the list elements have different data types. Please report this issue.\n\nCurrent value: {current_value}\nNew value: {value}')
            return False

        if not isinstance(value, type(current_value)):
            q.QMessageBox.critical(None, 'Bug',
                f'Refusing to set value because the data type is different (wanted {type(current_value)}, got {type(value)}). Please report this issue.\n\nCurrent value: {current_value}\nNew value: {value}')
            return False

        self.cdata[self.keys[row]] = value
        self.dataChanged.emit(index, index)
        return True

    def data(self, index: qc.QModelIndex, role: qc.Qt.ItemDataRole) -> qc.QVariant:
        col = index.column()
        row = index.row()
        item = self.cdata[self.keys[row]]

        if role == qc.Qt.UserRole:
            return item
        if role == qc.Qt.CheckStateRole:
            if col != ContainerModelColumn.Value or not isinstance(item, bool):
                return qc.QVariant()
            return qc.Qt.Checked if item else qc.Qt.Unchecked
        if role == qc.Qt.EditRole or role == qc.Qt.DisplayRole or role == qc.Qt.ToolTipRole:
            if col == ContainerModelColumn.Key:
                return self.keys[row]
            if col == ContainerModelColumn.Value:
                if isinstance(item, bool):
                    return qc.QVariant()
                if isinstance(item, ActorIdentifier):
                    return str(item)
                if isinstance(item, Argument):
                    return str(item)
                if isinstance(item, list):
                    return ', '.join(str(i) for i in item)
                return item
            if col == ContainerModelColumn.DataType:
                return util.get_container_value_type(item)

        return qc.QVariant()



import evfl.evfl
import evfl.actor
import evfl.event
import gzip
import io
import os
import typing
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtWidgets as q # type: ignore

def read_flow(path: str, flow: evfl.evfl.EventFlow):
    if path.endswith('.gz'):
        with gzip.open(path, 'rb') as f:
            flow.read(f.read())
    else:
        with open(path, 'rb') as f:
            flow.read(f.read())

def write_flow(path: str, flow: evfl.evfl.EventFlow):
    try:
        if path.endswith('.gz'):
            buf = io.BytesIO()
            flow.write(buf)
            with gzip.open(path + '.tmp', 'wb') as f:
                f.write(buf.getbuffer())
        else:
            with open(path + '.tmp', 'wb') as f:
                flow.write(f)
        os.replace(path + '.tmp', path)
    except:
        try:
            os.unlink(path + '.tmp')
        except FileNotFoundError:
            pass
        raise

def get_path(rel_path: str):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), rel_path)

def get_event_type(event: evfl.event.Event) -> str:
    if isinstance(event.data, evfl.event.ActionEvent):
        return 'Action'
    if isinstance(event.data, evfl.event.SwitchEvent):
        return 'Switch'
    if isinstance(event.data, evfl.event.ForkEvent):
        return 'Fork'
    if isinstance(event.data, evfl.event.JoinEvent):
        return 'Join'
    if isinstance(event.data, evfl.event.SubFlowEvent):
        return 'Sub flow'
    return '???'

def get_event_description(event: evfl.event.Event) -> str:
    if isinstance(event.data, evfl.event.ActionEvent):
        return f'{str(event.data.actor.v.identifier)}::{str(event.data.actor_action.v)}'
    if isinstance(event.data, evfl.event.SwitchEvent):
        return f'{str(event.data.actor.v.identifier)}::{str(event.data.actor_query.v)}'
    if isinstance(event.data, evfl.event.ForkEvent):
        return '–'
    if isinstance(event.data, evfl.event.JoinEvent):
        return '–'
    if isinstance(event.data, evfl.event.SubFlowEvent):
        return f'{event.data.res_flowchart_name}<{event.data.entry_point_name}>'
    return '???'

def get_event_next_summary(event: evfl.event.Event) -> str:
    if isinstance(event.data, evfl.event.ActionEvent) or isinstance(event.data, evfl.event.JoinEvent) or isinstance(event.data, evfl.event.SubFlowEvent):
        return event.data.nxt.v.name if event.data.nxt.v else '–'
    if isinstance(event.data, evfl.event.SwitchEvent):
        return f'{len(event.data.cases)} case(s)'
    if isinstance(event.data, evfl.event.ForkEvent):
        return f'{len(event.data.forks)} fork(s)'
    return '???'

def get_event_param_list(event: evfl.event.Event) -> typing.Dict[str, typing.Any]:
    if isinstance(event.data, evfl.event.ActionEvent) or isinstance(event.data, evfl.event.SwitchEvent) or isinstance(event.data, evfl.event.SubFlowEvent):
        if not event.data.params:
            return dict()
        return event.data.params.data
    return dict()

def get_event_full_description(event: evfl.event.Event) -> str:
    info = [event.name, get_event_type(event), get_event_description(event)]
    return ' - '.join([x for x in info if (x and x != '–')])

def get_container_value_type(value: typing.Any) -> str:
    if isinstance(value, bool):
        return 'bool'
    if isinstance(value, int):
        return 'int'
    if isinstance(value, float):
        return 'float'
    if isinstance(value, str):
        if isinstance(value, evfl.common.Argument):
            return 'argument'
        return 'string'
    if isinstance(value, evfl.common.ActorIdentifier):
        return 'actor id.'
    if isinstance(value, list):
        if isinstance(value[0], bool):
            return 'bools'
        if isinstance(value[0], int):
            return 'ints'
        if isinstance(value[0], float):
            return 'floats'
        if isinstance(value[0], str):
            return 'strings'
    return '???'

def is_valid_container_value_type(value: typing.Any) -> bool:
    return get_container_value_type(value) != '???'

def is_list_homogeneous(l: list) -> bool:
    for i in range(len(l) - 1):
        if type(l[i]) != type(l[i+1]):
            return False
    return True

def are_list_types_homogeneous_and_equal(list1: list, list2: list) -> bool:
    """Check whether both lists contain the same data type and only one."""
    if not (is_list_homogeneous(list1) and is_list_homogeneous(list2)):
        return False
    if list1 and list2 and type(list1[0]) != type(list2[0]):
        return False
    return True

def is_actor_in_use(events: typing.List[evfl.event.Event], actor: evfl.actor.Actor) -> bool:
    for event in events:
        if isinstance(event.data, evfl.event.ActionEvent) or isinstance(event.data, evfl.event.SwitchEvent):
            if event.data.actor.v == actor:
                return True
    return False

def is_actor_string_in_use(events: typing.List[evfl.event.Event], value: evfl.common.StringHolder) -> bool:
    for event in events:
        if isinstance(event.data, evfl.event.ActionEvent):
            if event.data.actor_action.v is value:
                return True
        if isinstance(event.data, evfl.event.SwitchEvent):
            if event.data.actor_query.v is value:
                return True
    return False

def connect_model_change_signals(model, flow_data, change_reason) -> None:
    def emit(*_) -> None:
        flow_data.flowDataChanged.emit(change_reason)

    model.dataChanged.connect(emit)
    model.rowsInserted.connect(emit)
    model.rowsRemoved.connect(emit)

class ItemEditorFactory(q.QItemEditorFactory):
    """Reimplements createEditor() to increase precision for doubles.

    Note: this MUST not be set as the default editor factory. Doing so will cause Qt to
    destruct the default factory and stop handling edit events for other simple data types."""
    def createEditor(self, userType, parent):
        if userType == qc.QVariant.Double:
            box = q.QDoubleSpinBox(parent)
            box.setDecimals(5)
            box.setSingleStep(0.1)
            box.setRange(-1000000, 1000000)
            return box
        return super().createEditor(userType, parent)

def set_view_delegate(view) -> None:
    delegate = q.QStyledItemDelegate()
    delegate.setItemEditorFactory(ItemEditorFactory())
    view.setItemDelegate(delegate)



# main.py
import chatgpt_events as cge

# Example of generating text
prompt = "Describe the historical significance of the Eiffel Tower."
print(cge.generate_text(prompt))

# Example of starting a chat conversation
initial_prompt = "Hello, can you help me understand quantum physics?"
text_response, chat_session = cge.start_chat_conversation(initial_prompt)
print(text_response)

# Continuing the conversation
follow_up = "What is the Heisenberg Uncertainty Principle?"
print(cge.send_chat_message(chat_session, follow_up))



import aamp
import byml
from enum import IntEnum
import functools
import os
from pathlib import Path
import typing
import oead

_rom_path: typing.Optional[Path] = None
def set_rom_path(p: typing.Optional[str]) -> None:
    if p:
        global _rom_path
        _rom_path = Path(p)

def _list_aiprog_files(path: Path):
    try:
        return path.glob('*.baiprog')
    except:
        return []

class AIProg:
    def __init__(self) -> None:
        self.actions: typing.Dict[str, str] = dict()
        self.queries: typing.Dict[str, str] = dict()

    def load_actor_aiprog(self, actor_name: str) -> bool:
        if not _rom_path:
            return False

        rel_roots = (
            '',
            'Pack/Bootup.pack/',
            'Pack/TitleBG.pack/',
            'Pack/RemainsWind.pack/',
            'Pack/RemainsElectric.pack/',
            'Pack/RemainsWater.pack/',
            'Pack/RemainsFire.pack/',
        )

        for rel_root in rel_roots:
            root = _rom_path / rel_root
            aiprog_dir = root/'Actor'/'Pack'/f'{actor_name}.sbactorpack'/'Actor'/'AIProgram'
            for path in _list_aiprog_files(aiprog_dir):
                with open(path, 'rb') as aiprog:
                    pio = aamp.Reader(aiprog.read()).parse()
                if self._do_load_actor_aiprog(pio):
                    return True

        return False

    def _do_load_actor_aiprog(self, aiprog: aamp.ParameterIO) -> bool:
        try:
            param_root = aiprog.list('param_root')
            action_list = param_root.list('Action')
            query_list = param_root.list('Query')

            for i in range(len(action_list.lists)):
                definition = action_list.list(f'Action_{i}').object('Def')
                class_name = definition.param('ClassName')
                try:
                    name = definition.param('Name')
                except KeyError:
                    name = class_name
                self.actions[name] = class_name

            for i in range(len(query_list.lists)):
                definition = query_list.list(f'Query_{i}').object('Def')
                class_name = definition.param('ClassName')
                try:
                    name = definition.param('Name')
                except KeyError:
                    name = class_name
                self.queries[name] = class_name

            return True

        except KeyError:
            return False

@functools.lru_cache(maxsize=50)
def load_aiprog(actor_name: str) -> typing.Optional[AIProg]:
    aiprog = AIProg()
    if not aiprog.load_actor_aiprog(actor_name):
        return None
    return aiprog

class AIType(IntEnum):
    Action = 0
    Query = 1

class AIParameter(typing.NamedTuple):
    name: str
    type: str
    default_value: typing.Any

    def get_default_value(self) -> typing.Any:
        if self.default_value is not None:
            return self.default_value
        if self.type == 'Bool':
            return False
        if self.type == 'Int':
            return 0
        if self.type == 'String':
            return ''
        if self.type == 'Float':
            return 0.0
        if self.type == 'Vec3':
            return [0.0, 0.0, 0.0]
        if self.type == 'AITreeVariablePointer' \
        or self.type == 'MesTransceiverId' \
        or self.type == 'BaseProcHandle' \
        or self.type == 'Actor':
            # These shouldn't ever appear in actions/queries.
            return None
        return '???'

class AIDef:
    def __init__(self) -> None:
        self._ai_defs: dict = dict()

    def _init_ai_defs(self) -> None:
        if self._ai_defs or not _rom_path:
            return

        raw_data = oead.yaz0.decompress(
            (_rom_path / 'Pack/Bootup.pack/Actor/AIDef/AIDef_Game.product.sbyml').read_bytes()
        )
        defs = byml.Byml(raw_data).parse()
        if isinstance(defs, dict):
            self._ai_defs = defs

    def get_parameters(self, ai_type: AIType, name: str) -> typing.List[AIParameter]:
        self._init_ai_defs()
        if not self._ai_defs:
            return []

        def map_to_ai_param(entry: dict) -> AIParameter:
            # {Name: XXXXX, Type: Int}
            # {Name: XXXXX, Type: Int, Value: 0}
            return AIParameter(name=entry['Name'], type=entry['Type'],
                default_value=entry.get('Value', None))

        try:
            if ai_type == AIType.Action:
                return [map_to_ai_param(x) for x in self._ai_defs['Actions'][name]['DynamicInstParams']]
            if ai_type == AIType.Query:
                return [map_to_ai_param(x) for x in self._ai_defs['Querys'][name]['DynamicInstParams']]
            return []
        except:
            return []

ai_def_instance = AIDef()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="eventeditor-GPT",
    version="1.0.2",  # Static version number
    author="leoetlino",
    author_email="leo@leolam.fr",
    description="Event editor for Breath of the Wild and Tears of the Kingdom",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/leoetlino/event-editor",
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
    ],
    include_package_data=True,
    install_requires=[
        'evfl~=1.1',
        'pyqt5-sip~=12.7',
        'pyqtwebengine~=5.14',
        'PyYAML~=5.1',
        'aamp~=1.0',
        'byml~=2.0',
        'oead~=1.1',
    ],
    python_requires='>=3.6',
    entry_points={
        'gui_scripts': [
            'eventeditor = eventeditor.__main__:main'
        ],
    },
)



from enum import IntEnum, auto
import typing

from eventeditor.event_chooser_dialog import EventChooserDialog
from eventeditor.flow_data import FlowData, FlowDataChangeReason
import eventeditor.util as util
from evfl import Event
from evfl.common import RequiredIndex
import PyQt5.QtCore as qc # type: ignore
import PyQt5.QtWidgets as q # type: ignore

_Cases = typing.Dict[int, RequiredIndex[Event]]
_Forks = typing.List[RequiredIndex[Event]]

_PLACEHOLDER_EVENT = Event()
_PLACEHOLDER_EVENT.name = '<placeholder>'

class EventBranchEditorTableView(q.QTableView):
    chooserEventDoubleClicked = qc.pyqtSignal(int)
    chooserSelectSignal = qc.pyqtSignal(int)
    actionProhibitionChanged = qc.pyqtSignal(bool)

    def __init__(self, parent, flow_data: FlowData) -> None:
        super().__init__(parent)
        self.flow_data = flow_data

    def edit(self, index: qc.QModelIndex, trigger, event) -> bool:
        if not (self.editTriggers() & trigger) or not index.isValid() or not (index.flags() & (qc.Qt.ItemIsEditable | qc.Qt.ItemIsUserCheckable)):
            return False

        data = index.data(qc.Qt.EditRole)
        if isinstance(data, Event):
            self.actionProhibitionChanged.emit(True)
            dialog = EventChooserDialog(self, self.flow_data, enable_ctx_menu=False)
            dialog.show()
            self.chooserSelectSignal.connect(dialog.event_view.selectEvent)
            try:
                dialog.event_view.selectEvent(self.flow_data.flow.flowchart.events.index(data))
            except ValueError:
                pass
            dialog.event_view.jumpToFlowchartRequested.connect(self.chooserEventDoubleClicked)
            dialog.finished.connect(lambda: self.actionProhibitionChanged.emit(False))
            def onChooserAccept():
                selected_event = dialog.getSelectedEvent()
                self.model().setData(index, selected_event, qc.Qt.EditRole)
            dialog.accepted.connect(onChooserAccept)
            return False

        return super().edit(index, trigger, event)

class SwitchCase:
    __slots__ = ('value', 'event')
    def __init__(self, value: int, event: Event) -> None:
        self.value = value
        self.event = event

class SwitchCaseModelColumn(IntEnum):
    Value = 0
    Event = auto()

class SwitchCaseModel(qc.QAbstractTableModel):
    def __init__(self, parent, cases: _Cases) -> None:
        super().__init__(parent)
        self.setCases(cases)

    def isValid(self) -> bool:
        for i in range(len(self.l)):
            if self.l[i].event is _PLACEHOLDER_EVENT:
                return False
            if i != len(self.l) - 1 and self.l[i].value == self.l[i + 1].value:
                return False
        return True

    def updateCaseDict(self, case_dict: _Cases) -> None:
        case_dict.clear()
        for case in sorted(self.l, key=lambda c: c.value):
            idx: RequiredIndex[Event] = RequiredIndex()
            idx.v = case.event
            case_dict[case.value] = idx

    def setCases(self, cases: _Cases) -> None:
        self.beginResetModel()
        self.l: typing.List[SwitchCase] = []
        for value, event_idx in cases.items():
            self.l.append(SwitchCase(value, event_idx.v))
        self.endResetModel()

    def columnCount(self, parent):
        return len(SwitchCaseModelColumn)

    def rowCount(self, parent):
        return len(self.l)

    def flags(self, index: qc.QModelIndex) -> qc.Qt.ItemFlags:
        return qc.Qt.ItemIsEditable | super().flags(index)

    def hasCaseValue(self, value: int) -> bool:
        return any(case.value == value for case in self.l)

    def appendCase(self, case: SwitchCase) -> bool:
        if self.hasCaseValue(case.value):
            q.QMessageBox.critical(None, 'Invalid data', f'Value {case.value} is already handled by another case.')
            return False
        self.beginInsertRows(qc.QModelIndex(), len(self.l), len(self.l))
        self.l.append(case)
        self.endInsertRows()
        return True

    def removeCase(self, row: int) -> bool:
        self.beginRemoveRows(qc.QModelIndex(), row, row)
        self.l.pop(row)
        self.endRemoveRows()
        return True

    def setData(self, index: qc.QModelIndex, value, role) -> bool:
        if role != qc.Qt.EditRole:
            return False
        col = index.column()
        row = index.row()
        if self.l[row].value == value:
            return True
        if col == SwitchCaseModelColumn.Value and isinstance(value, int):
            if self.hasCaseValue(value):
                q.QMessageBox.critical(None, 'Invalid data', f'Value {value} is already handled by another case.')
                return False
            self.l[row].value = value
        elif col == SwitchCaseModelColumn.Event and isinstance(value, Event):
            self.l[row].event = value
        else:
            return False
        self.dataChanged.emit(index, index)
        return True

    def data(self, index: qc.QModelIndex, role) -> typing.Any:
        col = index.column()
        row = index.row()
        if role == qc.Qt.UserRole:
            return self.l[row]
        if role == qc.Qt.DisplayRole or role == qc.Qt.ToolTipRole or role == qc.Qt.EditRole:
            if col == SwitchCaseModelColumn.Value:
                return self.l[row].value
            if col == SwitchCaseModelColumn.Event:
                event = self.l[row].event
                if role == qc.Qt.EditRole:
                    return event
                if event is _PLACEHOLDER_EVENT:
                    return '<placeholder>'
                return f'{util.get_event_full_description(event)}'
        return qc.QVariant()

    def headerData(self, section, orientation, role) -> qc.QVariant:
        if role != qc.Qt.DisplayRole:
            return qc.QVariant()
        if section == SwitchCaseModelColumn.Value:
            return 'Value'
        if section == SwitchCaseModelColumn.Event:
            return 'Event'
        return 'Unknown'

class SwitchEventEditDialog(q.QDialog):
    chooserEventDoubleClicked = qc.pyqtSignal(int)
    chooserSelectSignal = qc.pyqtSignal(int)

    def __init__(self, parent, cases: _Cases, flow_data: FlowData) -> None:
        super().__init__(parent)
        self.setWindowTitle('Edit switch event')
        self.setMinimumWidth(600)
        self.flow_data = flow_data
        self.orig_cases = cases
        self.model = SwitchCaseModel(self, cases)

        self.tview = EventBranchEditorTableView(None, self.flow_data)
        self.tview.setModel(self.model)
        self.tview.verticalHeader().hide()
        self.tview.setSelectionBehavior(q.QAbstractItemView.SelectRows)
        self.tview.setSelectionMode(q.QAbstractItemView.SingleSelection)
        self.tview.horizontalHeader().setMinimumSectionSize(80)
        self.tview.horizontalHeader().setSectionResizeMode(q.QHeaderView.ResizeToContents)
        self.tview.horizontalHeader().setSectionResizeMode(1, q.QHeaderView.Stretch)
        self.tview.setContextMenuPolicy(qc.Qt.CustomContextMenu)
        self.tview.customContextMenuRequested.connect(self.onContextMenu)
        self.tview.chooserEventDoubleClicked.connect(self.chooserEventDoubleClicked)
        self.chooserSelectSignal.connect(self.tview.chooserSelectSignal)

        add_btn_box = q.QHBoxLayout()
        add_btn = q.QPushButton('Add case')
        add_btn.clicked.connect(self.addCase)
        add_btn_box.addStretch()
        add_btn_box.addWidget(add_btn)

        btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Save | q.QDialogButtonBox.Cancel);
        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)
        layout = q.QVBoxLayout(self)
        layout.addWidget(self.tview)
        layout.addLayout(add_btn_box)
        layout.addWidget(btn_box)

        self.tview.actionProhibitionChanged.connect(lambda v: self.setEnabled(not v))

    def closeEvent(self, event):
        if not self.isEnabled():
            event.ignore()

    def accept(self) -> None:
        if not self.model.isValid():
            q.QMessageBox.critical(self, 'Invalid data', 'Please ensure there are no placeholder or duplicate cases left.')
            return

        self.model.updateCaseDict(self.orig_cases)
        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.Events)
        super().accept()

    def addCase(self) -> None:
        value, ok = q.QInputDialog.getInt(self, 'Add new switch case', 'Please enter the case value:')
        if not ok:
            return
        self.model.appendCase(SwitchCase(value, _PLACEHOLDER_EVENT))

    def onContextMenu(self, pos) -> None:
        smodel = self.tview.selectionModel()
        if not smodel.hasSelection():
            return
        sidx = smodel.selectedRows()[0]
        menu = q.QMenu()
        menu.addAction('Remove', lambda: self.model.removeCase(sidx.row()))
        menu.exec_(self.sender().viewport().mapToGlobal(pos))

class ForkEventModel(qc.QAbstractListModel):
    def __init__(self, parent, forks: _Forks) -> None:
        super().__init__(parent)
        self.setForks(forks)

    def isValid(self) -> bool:
        for i in range(len(self.l)):
            if self.l[i] is _PLACEHOLDER_EVENT:
                return False
            if i != len(self.l) - 1 and self.l[i] == self.l[i + 1]:
                return False
        return True

    def updateForkList(self, fork_list: _Forks) -> None:
        fork_list.clear()
        for fork in self.l:
            idx: RequiredIndex[Event] = RequiredIndex()
            idx.v = fork
            fork_list.append(idx)

    def setForks(self, forks: _Forks) -> None:
        self.beginResetModel()
        self.l: typing.List[Event] = []
        for event_idx in forks:
            self.l.append(event_idx.v)
        self.endResetModel()

    def rowCount(self, parent):
        return len(self.l)

    def flags(self, index: qc.QModelIndex) -> qc.Qt.ItemFlags:
        return qc.Qt.ItemIsEditable | super().flags(index)

    def hasFork(self, fork: Event) -> bool:
        return any(f == fork for f in self.l)

    def appendFork(self, fork: Event) -> bool:
        if self.hasFork(fork):
            q.QMessageBox.critical(None, 'Invalid data', f'{fork.name} is already a fork branch.')
            return False
        self.beginInsertRows(qc.QModelIndex(), len(self.l), len(self.l))
        self.l.append(fork)
        self.endInsertRows()
        return True

    def removeCase(self, row: int) -> bool:
        self.beginRemoveRows(qc.QModelIndex(), row, row)
        self.l.pop(row)
        self.endRemoveRows()
        return True

    def setData(self, index: qc.QModelIndex, value, role) -> bool:
        if role != qc.Qt.EditRole:
            return False
        if self.l[index.row()] == value:
            return True
        if isinstance(value, Event):
            self.l[index.row()] = value
            self.dataChanged.emit(index, index)
            return True
        return False

    def data(self, index: qc.QModelIndex, role) -> typing.Any:
        row = index.row()
        if role == qc.Qt.UserRole:
            return self.l[row]
        if role == qc.Qt.DisplayRole or role == qc.Qt.ToolTipRole or role == qc.Qt.EditRole:
            event = self.l[row]
            if role == qc.Qt.EditRole:
                return event
            if event is _PLACEHOLDER_EVENT:
                return '<placeholder>'
            return f'{util.get_event_full_description(event)}'
        return qc.QVariant()

# This is pretty similar to the switch edit dialog.
# TODO: refactor and deduplicate this
class ForkEventEditDialog(q.QDialog):
    chooserEventDoubleClicked = qc.pyqtSignal(int)
    chooserSelectSignal = qc.pyqtSignal(int)

    def __init__(self, parent, forks: _Forks, flow_data: FlowData) -> None:
        super().__init__(parent)
        self.setWindowTitle('Edit fork event')
        self.setMinimumWidth(600)
        self.flow_data = flow_data
        self.orig_forks = forks
        self.model = ForkEventModel(self, forks)

        self.tview = EventBranchEditorTableView(None, self.flow_data)
        self.tview.setModel(self.model)
        self.tview.horizontalHeader().hide()
        self.tview.verticalHeader().hide()
        self.tview.setSelectionBehavior(q.QAbstractItemView.SelectRows)
        self.tview.setSelectionMode(q.QAbstractItemView.SingleSelection)
        self.tview.horizontalHeader().setSectionResizeMode(q.QHeaderView.Stretch)
        self.tview.setContextMenuPolicy(qc.Qt.CustomContextMenu)
        self.tview.customContextMenuRequested.connect(self.onContextMenu)
        self.tview.chooserEventDoubleClicked.connect(self.chooserEventDoubleClicked)
        self.chooserSelectSignal.connect(self.tview.chooserSelectSignal)

        add_btn_box = q.QHBoxLayout()
        add_btn = q.QPushButton('Add fork')
        add_btn.clicked.connect(self.addFork)
        add_btn_box.addStretch()
        add_btn_box.addWidget(add_btn)

        btn_box = q.QDialogButtonBox(q.QDialogButtonBox.Save | q.QDialogButtonBox.Cancel);
        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)
        layout = q.QVBoxLayout(self)
        layout.addWidget(self.tview)
        layout.addLayout(add_btn_box)
        layout.addWidget(btn_box)

        self.tview.actionProhibitionChanged.connect(lambda v: self.setEnabled(not v))

    def closeEvent(self, event):
        if not self.isEnabled():
            event.ignore()

    def accept(self) -> None:
        if not self.model.isValid():
            q.QMessageBox.critical(self, 'Invalid data', 'Please ensure there are no placeholder or duplicate forks left.')
            return

        if self.model.rowCount(qc.QModelIndex()) == 0:
            q.QMessageBox.critical(self, 'Invalid data', 'Fork events must always have at least one branch.')
            return

        self.model.updateForkList(self.orig_forks)
        self.flow_data.flowDataChanged.emit(FlowDataChangeReason.Events)
        super().accept()

    def addFork(self) -> None:
        self.model.appendFork(_PLACEHOLDER_EVENT)

    def onContextMenu(self, pos) -> None:
        smodel = self.tview.selectionModel()
        if not smodel.hasSelection():
            return
        sidx = smodel.selectedRows()[0]
        menu = q.QMenu()
        menu.addAction('Remove', lambda: self.model.removeCase(sidx.row()))
        menu.exec_(self.sender().viewport().mapToGlobal(pos))
